/*
data structures which describe registers on the Xilinx
Complex Programmable Logic Device (CPLD) attached between
the AMCC PCI bus interface and the RX FPGA, TX FPGA, and
AIM optics module.



Note 1: these data structures use bit fields, which are compiler-
and Central Processing Unit (CPU) specific, so you will probably
have to rewrite these data structures if you port to another
CPU or compiler

On the Intel CPU's using the Borland C++ 4.5 compiler, bits are
assigned from least significant to most significant within a word,
and less significant bytes of a word are stored at lower memory
addresses

No padding is inserted, and bitfields can straddle word boundaries

The width of each bitfield must be from 1 to 16 (see page 62 of the
Borland C++ version 4.5 Programmer's Guide).



Note 2: due to the use of anonymous unions, this header can only be
included from C++ code, not C code



Note 3: definitions of readability/writability abbreviations:
    W/O     writable only
    R/O     readable only
    W/R     writable and readable
    CW/R    readable, but writable only after
            CWRE bit set in Control_register
*/



#ifndef _CPLD_H
#define _CPLD_H



#include "../types/types.h"



struct CpldControl1_register
{
    Bit16 ProgramNot        : 1;            // R/W
    Bit16 Done              : 1;            // R/O
    Bit16 Reserved1         : 8;
    Bit16 ConfigErrorNot    : 1;            // R/O
    Bit16 Reserved2         : 5;
};



// merged registers
//
struct Cpld_registers
{
    union
    {
        CpldControl1_register Control1;
        Bit32 dword_0;
    };

    Bit32 Reserved[ 254 ];
};



#endif      // not included yet
