#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "..\types\types.h"             // Bit32 Bit8



// manifest constants
//
#define DISPLAY_MODULUS                 1000
#define IN_BUFFER_SIZE                  516             // 2 type + 2 length + 2*255 max length value + possible newline + terminating NUL
#define OUT_BUFFER_SIZE                 255             // 255 max length value - min 2 byte length




// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define SEGREGATE_BYTES                 FALSE           // TRUE screws up at whitespace!



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *outfile = argv[ 2 ];
    FILE *in_handle;
    FILE *out_handle;
    int result;
    char *result_string;
    char in_buffer[ IN_BUFFER_SIZE ];
    Bit8 out_buffer[ OUT_BUFFER_SIZE ];
    Bit32 line_count = 0;
    Bit32 valid_srec_count = 0;
    Bit32 byte_count = 0;
    int one_byte;                       // mem corruption if not wide enough!

    if( argc != 3 )
    {
        fprintf( stderr, "converts a file with Motorola S-records into binary bytes,\n" );
        fprintf( stderr, "ignoring all lines that do not start with 'S'\n\n" );
        fprintf( stderr, "usage: %s infile outfile\n", program );
        return 1;
    }

    in_handle = fopen( infile, "r" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    out_handle = fopen( outfile, "wb" );
    if( out_handle == NULL )
    {
        fprintf( stderr, "error %d opening output file\n", errno );
        return 3;
    }

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        // try to read a whole line (single S-record) from the input file
        //
        result_string = fgets( in_buffer, IN_BUFFER_SIZE, in_handle );
        if( result_string == NULL )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file\n", errno );

            break;
        }

        // count the lines read
        //
        line_count++;

        // check that the first character on the line is "S"
        //
        if( in_buffer[0] != 'S' )
            continue;

        // compare the second character on the line
        //
        switch( in_buffer[1] )
        {
            default:
                fprintf( stderr, "warning: unknown S-record type '%c' on line %d ignored\n", in_buffer[1], line_count );
                continue;

            case '0':
            case '1':
            case '2':
            case '3':
            case '5':
            case '7':
            case '8':
            case '9':
        }

        // convert the length to binary
        //

        // do something with the address field
        // (probably discard it)
        //

        // subtract the length of the address field
        // from the total we got in the record header
        //

        // for each byte in the code/data field...
        //

            // try to read a hex-encoded byte from the string
            //
            result = sscanf( in_string+2*offset, "%2X", &one_byte );
            if( result != 1 )
            {
                fprintf( stderr, "cannot convert string '%.2s' as hex\n", in_string+2*offset );
                break;
            }

            // update the running checksum
            //

            // no need to swap the bits in the byte lsb-for-msb
            // because that was already done by MAKEPROM Xilinx
            // program (.HEX files need this, though)
            //


            out_buffer[ offset ] = one_byte;

        // compare calculated checksum to the one in the record
        //

        // make sure there is nothing else on the line,
        // other than a newline (fgets copied the newline)
        //
        // maybe use sscanf() with whitespace format string???
        //

        // copy the output buffer to the output file
        //
        result = fwrite( ??? );
        if( result == -1 )
        {
            fprintf( stderr, "error %d writing output file\n", errno );
            break;
        }

        // count valid S-records read
        //
        valid_srec_count++;

        // count the bytes output
        //
        byte_count += length;

        // show some status to the user (occasionally)
        //
        if( byte_count % DISPLAY_MODULUS == 0 )
            fprintf( stderr, "%lu ", byte_count );

    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    // show statistics
    //
    fprintf( stdout, "%lu bytes output\n", byte_count );

    fclose( out_handle );
    fclose( in_handle );

    // return successfully
    //
    return 0;
}

