#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../intel/timestmp.h"          // class PentiumClock
#include "../apptel/driver.h"           // class AtmCardRunning
#include "../apptel/intrtest.h"         // our prototypes and errors
#include "../dpmi32/phys.h"             // flat_to_linear()
#include "../dpmi32/flat.h"             // flat_init() flat_exit()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    int want_write;
    int result;
    const char *program_name = argv[0];
    double max_time;
    PentiumClock start, alarm, now;
    AtmCardRunning *card;
    Bit8 *data;
    BufferBlock *block;
    unsigned processed_blocks = 0;

    // parse command-line arguments
    //
    if( argc != 3 )
    {
        fprintf( stderr, "tests interrupt handling for POINT card\n\n" );
        fprintf( stderr, "usage: %s cardindex max-time-in-seconds\n", program_name );
        return ERROR_POINT_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    max_time = strtod( argv[2], NULL );
    fprintf( stdout, "waiting a maximum of %g seconds\n", max_time );

    // we must take care of initializing flat memory
    // before the BufferQueue constructor gets called
    //
    flat_init();

    // allocate all remaining flat memory to card buffer pool
    //
    // we must do this AFTER statistics initialize, so they
    // get first crack at flat memory, in case they are using
    // it instead of malloc()
    //
    all_buffers = new BufferQueue;
    if( all_buffers == NULL )
    {
        printf( "fatal error: cannot allocate BufferQueue\n" );
        return ERROR_POINT_CANNOT_ALLOC;
    }

    // say where the global buffers are
    //
    printf( "global control blocks are at 0x%p\n", &all_buffers );

    data = all_buffers->get_start()->buffers;
    printf( "global data buffers are at   log/lin/phy"
        " 0x%4.4X:%8.8lX 0x%8.8lX 0x%8.8lX\n", _DS, data,
        flat_to_linear(data), flat_to_physical(data) );

    printf( "SW_RINGS=%d\n", SW_RINGS );

    printf( "COMPL_RING_ENTRIES=%d\n", COMPL_RING_ENTRIES );

    card = new AtmCardRunning( card_index );
    if( !card )
        return ERROR_POINT_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    // now tell the card to do something useful
    //
    result = card->card_init( AtmCardRunning::PassIdleCells, Equipment, TRUE, FALSE );
    if( result )
        return result;

    printf( "test is running\n" );

    // wait until an interrupt occurs
    // or the specified amount of time elapses
    //
    alarm.put_seconds( max_time );
    start.set();
    alarm += start;
    do
    {
        // read a buffer out of the filled list
        //
        result = all_buffers->read( &block );
        if( !result )
        {
            // count it
            //
            processed_blocks++;

            // free it
            //
            result = all_buffers->free( block );
            if( result )
            {
                printf( "error %d from BufferQueue::free( %p )\n", result, block );
                break;
            }
        }

        now.set();
    } while( all_buffers->allocate_failure_count<1 && now<alarm );

    printf( "test done\n" );

    printf( "foreground loop 'processed' %u blocks\n", processed_blocks );
    printf( "BufferQueue allocation failed %d times\n", all_buffers->allocate_failure_count );

    delete card;
    delete all_buffers;

    flat_exit();

    if( now >= alarm )
    {
        fprintf( stdout, "buffers were never exhausted during %g second run\n", max_time );
        return 0;
    }

    now -= start;
    fprintf( stdout, "buffers exhausted after %g milliseconds\n", 1000*now.get_seconds() );

    return 0;
}
