#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/nvram.h"            // nvram_init_stats() nvram_print_stats() nvram_read_or_write()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    Bit16 nvram_address;
    Bit16 nvram_length;
    Bit16 count;
    Bit8 nvram_data;
    int result;
    unsigned line;
    const char *program_name = argv[0];
    AtmCard *card;

    // parse command-line arguments
    //
    if( argc != 4 )
    {
        fprintf( stderr, "dumps a range of NVRAM bytes from the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex address length-in-bytes\n", program_name );
        return ERROR_NVRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    nvram_address = strtoul( argv[2], NULL, 0 );
    nvram_length = strtoul( argv[3], NULL, 0 );

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_NVRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    nvram_init_stats();

    printf( "dumping POINT NVRAM values starting at address 0x%.4X, word length 0x%.4X\n", nvram_address, nvram_length );

    for( count=0; count<nvram_length; count+=16 )
    {
        // this line assumes NVRAM addresses have 16 bits or less
        //
        printf( "%.4X  ", nvram_address+count );

        for( line=0; line<16 && count+line<nvram_length; line++ )
        {
            result = nvram_read_or_write( card->amcc, nvram_address+count+line, FALSE, &nvram_data );
            if( result )
                break;
            printf( " %.2X", nvram_data );
        }

        printf( "\n" );
    }
    printf( "\n" );

    nvram_print_stats();

    delete card;

    if( result )
    {
        fprintf( stderr, "error %d from nvram_read_or_write() on %uth word\n", result, count );
        return result;
    }

    return 0;
}
