#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/sram.h"             // sram_init_stats() sram_print_stats() sram_read_or_write()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    Bit32 sram_address;
    Bit32 sram_length;
    Bit32 count;
    Bit32 sram_data;
    int result;
    const char *program_name = argv[0];
    AtmCard *card;

    // parse command-line arguments
    //
    if( argc != 5 )
    {
        fprintf( stderr, "fills a single value in a range of SRAM words on the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex address length-in-words value\n", program_name );
        return ERROR_SRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    sram_address = strtoul( argv[2], NULL, 0 );
    sram_length = strtoul( argv[3], NULL, 0 );
    sram_data = strtoul( argv[4], NULL, 0 );

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_SRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    sram_init_stats();

    printf( "filling POINT SRAM with value 0x%.8X starting at address 0x%.8X, length 0x%.8X words\n", sram_data, sram_address, sram_length );

    for( count=0; count<sram_length; count++ )
    {
        result = sram_read_or_write( &card->both_fpgas->Receive, sram_address+count, TRUE, &sram_data );
        if( result )
            break;
    }

    sram_print_stats();

    delete card;

    if( result )
    {
        fprintf( stderr, "error %d from sram_read_or_write() on %uth word\n", result, count );
        return result;
    }

    return 0;
}
