#include <stdio.h>                      // printf()
#include <stdlib.h>                     // atoi() strtoul()
#include "../apptel/point.h"            // class AtmCard
#include "../apptel/sram.h"             // sram_init_stats() sram_print_stats() sram_read_or_write()



int main( int argc, char *argv[] )
{
    Bit16 card_index;
    Bit32 sram_address;
    Bit32 sram_data;
    int want_write;
    int result;
    const char *program_name = argv[0];
    AtmCard *card;

    // parse command-line arguments
    //
    if( argc<3 || argc>4 )
    {
        fprintf( stderr, "reads or writes single SRAM words on the Applied Telecom PCI card\n\n" );
        fprintf( stderr, "usage: %s cardindex address {new-value}\n", program_name );
        return ERROR_SRAM_MISSING_ARGS;
    }

    card_index = atoi( argv[1] );
    sram_address = strtoul( argv[2], NULL, 0 );

    if( argc == 4 )
    {
        sram_data = strtoul( argv[3], NULL, 0 );
        want_write = TRUE;
    }
    else
        want_write = FALSE;

    card = new AtmCard( card_index );
    if( !card )
        return ERROR_SRAM_CANNOT_ALLOC;
    if( card->init_result )
        return card->init_result;

    sram_init_stats();

    result = sram_read_or_write( &card->both_fpgas->Receive, sram_address, want_write, &sram_data );

    sram_print_stats();

    delete card;

    if( result )
    {
        fprintf( stderr, "error %d from sram_read_or_write()\n", result );
        return result;
    }

    if( want_write )
        printf( "wrote new value 0x%.8X at address 0x%.8X\n", sram_data, sram_address );
    else
        printf( "read old value 0x%.8X at address 0x%.8X\n", sram_data, sram_address );

    return 0;
}
