#include <stdio.h>              // printf()
#include <stdlib.h>             // exit()
#include "phys.h"



typedef void (*writer_t)( void *pointer, Bit32 value );



// this set of 3 functions takes care of touching
// only the address and width that is wanted by
// the user
//
// no masking is necessary because C promotion
// will zero out the high bits if necessary
//
void write32( void *pointer, Bit32 value )
{
    * (Bit32 *) pointer = value;
}



void write16( void *pointer, Bit32 value )
{
    * (Bit16 *) pointer = value;
}



void write8( void *pointer, Bit32 value )
{
    * (Bit8 *) pointer = value;
}



int main( int argc, char *argv[] )
{
    Bit32 handle, start, line;
    Signed32 length, length_remaining;
    Bit32 value;
    int width;
    const char *program_name = argv[0];
    Bit8 *pointer, *address;
    writer_t writer;

    // parse command-line arguments
    //
    if( argc != 5 )
    {
        fprintf( stderr, "usage: %s phys-address bit-width length value\n", program_name );
        return 1;
    }
    start  = strtoul( argv[1], NULL, 0 );
    length = strtoul( argv[3], NULL, 0 );
    width  = atoi( argv[2] );
    value  = strtoul( argv[4], NULL, 0 );

    switch( width )
    {
        default:
            fprintf( stderr, "Sorry, you can only use widths of 8, 16, or 32 bits.\n" );
            return 3;

        case 8:
            writer = write8;
            break;

        case 16:
            writer = write16;
            break;

        case 32:
            writer = write32;
            break;
    }

    // map one or more pages of memory starting at the
    // physical address that the user specified
    //
    pointer = (Bit8 *) map_physical( start, length, &handle );
    if( pointer == NULL )
    {
        printf( "Sorry, cannot map %d bytes at physical address 0x%.8X to a flat pointer\n", length, start );
        return 2;
    }

    printf( "filling memory with %d-bit value 0x%.*X starting at address 0x%.8X, length 0x%.8X\n", width, width/4, value, start, length );

    for( address=pointer,length_remaining=length; length_remaining>0; address+=16,length_remaining-=16 )
    {
        for( line=0; line<16 && line<length_remaining; line+=width/8 )
        {
            writer( address + line, value );
        }
    }

    return 0;
}
