/*
 * Copyright (c) 1993 by Fore Systems, Inc.
 */
#ifndef _board_h
#define _board_h

#ifndef lint
static char _board_h_rcsid[] = "@(#)$Id: board.h,v 1.72 1995/03/07 16:57:29 ghoti Exp $ FSI";
#endif /* lint */

/*
 * AIX expects all enum values to be signed only
 */
enum boot_status {
	cold_start     = -1071792099,	/* 0xc01dc01d */
	self_test_ok   = 0x02201958,
	self_test_fail = -1380262995,	/* 0xadbadbad */
	cp_running     = -837681427,	/* 0xce11feed - application is ready */
	mon960_too_big = 0x10aded00
};
typedef enum boot_status Boot_status;

/*
 * Structure shared between host and CP,
 * found in low CP memory.
 */
struct mon960com {
	Soft_UART    uart;     /* accessed by mon960 and ncomm */
	Boot_status  bstat;
	u_int        app_base; /* application base offset */
	int			 version;	/* mon960 version */
};
typedef volatile struct mon960com Mon960com; 

#define PROM_SIZE		(32 * 1024)

#define PROM_OFFSET     0x00000000
#define CTRL_OFFSET     0x00800000
#define RAM_OFFSET      0x01000000

/*********              START   BUS   ASIC  RELATED DEFINITIONS   **********/
        /*
         * This section is intended to be used for all bus asics. Currently
         * SDC(sbus dma controller) and EDC ( eisa dma controller ) are
         * defined. Due to the read-modify-write operation required for the
         * HCR(host control register) operations, the #defines and semantics
         * are seperated. Also there may be a need forboth 200 and 200E series
         * boards to be present in different slots.
         */

/*      Sbus DMA Controller (SDC)

+------------------------------------------------------------------------------+
| Host Control Register                                                        |
|Bit:  |   8   |   7   |   6   |   5   |   4   |   3   |   2   |   1   |   0   |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
|Read  | SBus  | Test  | INFF  | ESP   | SBus  |OutFF  | i960  | i960  | Board |
|      |Interpt| Mode  |AlFull | Hold  |IEnable|AlFull |FAIL ST|HLDLOCK| Reset |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
|Write | XXXX  | XXXX  | XXXX  | XXXX  | SBus  | Clear | Slave | i960  | Board |
|      | XXXX  | XXXX  | XXXX  | XXXX  |IEnable|Bus Int|Interpt|HLDLOCK| Reset |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
*/


/* SDC Host Control Register(HCR) Write definitions */

#define SBA200E_BOARD_RESET_SET         (1 << 0) /* Resets both i960 and ESP */
#define SBA200E_BOARD_RESET_CLEAR       0x01e    /* Clear reset by writing a 0 */
#define SBA200E_HOLD_960_BUS            (1 << 1) /* Effectively makes i960 idle */
#define SBA200E_RELEASE_960_BUS         0x01d
#define SBA200E_INTERRUPT_960           (1 << 2) /* SDC interrupting i960 */
#define SBA200E_CLEAR_INTERRUPT         (1 << 3) /* Clears pending SBus interrupt */
#define SBA200E_INTERRUPT_ENABLE        (1 << 4) /* Allows SBus intrpts (intr mask)*/
#define SBA200E_SBUS_INTERRUPT_DISABLE  0x0f

/* Read Definitions */

#define SBA200E_I960_SELFTEST_FAIL      (1 << 2)  /* Did the i960 fail selftest? */
#define SBA200E_OFIFO_ALMOST_FULL       (1 << 3)  /* Is the OFifo almost full? */
#define SBA200E_ESP_HOLD                (1 << 5)  /* Is ESP holding bus? */
#define SBA200E_IFIFO_ALMOST_FULL       (1 << 6)  /* Is IN fifo almost full? */
#define SBA200E_TEST_MODE               (1 << 7)  /* Are we in test mode? */
#define SBA200E_SBUS_INTERRUPT          (1 << 8)  /* Is SBus intrpt pending? */


/*      EISA    BUS INTERFACE ASIC ( EBI)

+------------------------------------------------------------------------------+
| Host Control Register                                                        |
|Bit:  |   8   |   7   |   6   |   5   |   4   |   3   |   2   |   1   |   0   |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
|Read  | XXXX  | Test  | InFIFO| ESP   |OutFifo| Hold  | i960  | LOC   | Board |
|      | XXXX  | Mode  |AlmostF| Hold  | Hold  | ACK   | FAIL  | Hold  | Reset |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
|Write | XXXX  | XXXX  | XXXX  | XXXX  | Clear |Interpt|Interpt| LOC   | Board |
|      | XXXX  | XXXX  | XXXX  | XXXX  | Intr  | i960A | i960B | Hold  | Reset |
|------+-------+-------+-------+-------+-------+-------+-------+-------+-------+
*/
/* EBI Host Control Register(HCR) Write definitions */

#define ESA200E_BOARD_RESET_SET         (1 << 0) /* Resets both i960 and ESP */
#define ESA200E_BOARD_RESET_CLEAR       0x01e    /* Clear reset by writing a 0 */
#define ESA200E_HOLD_960_BUS            (1 << 1) /* Effectively makes i960 idle */
#define ESA200E_RELEASE_960_BUS         0x01d
#define ESA200E_INTERRUPT_960A          (1 << 2) /* EBI interrupting i960 */
#define ESA200E_INTERRUPT_960B          (1 << 3) /* EBI interrupting i960 */
#define ESA200E_CLEAR_INTERRUPT         (1 << 4) /* Clears pending EISA interrupt */

/* Read Definitions */

#define ESA200E_I960_FAIL      		(1 << 2)  /* Did the i960 fail selftest? */
#define ESA200E_I960_HOLD_ACK		(1 << 3)
#define ESA200E_OFIFO_ALMOST_FULL       (1 << 4)  /* Is the OFifo almost full? */
#define ESA200E_ESP_HOLD                (1 << 5)  /* Is ESP holding bus? */
#define ESA200E_IFIFO_ALMOST_FULL       (1 << 6)  /* Is IN fifo almost full? */
#define ESA200E_TEST_MODE               (1 << 7)  /* Are we in test mode? */


#ifdef SDC_ASIC

#define HCR_MASK                        0x13

#define E_BOARD_RESET                   SBA200E_BOARD_RESET_SET
#define E_CLEAR_BOARD_RESET             SBA200E_BOARD_RESET_CLEAR
#define E_HOLD_960_BUS                  SBA200E_HOLD_960_BUS
#define E_CLEAR_HOLD_960_BUS            SBA200E_RELEASE_960_BUS
#define E_INTERRUPT_960                 SBA200E_INTERRUPT_960
#define E_CLEAR_INTERRUPT               SBA200E_CLEAR_INTERRUPT
#define E_ENABLE_BUS_INTERRUPT          SBA200E_INTERRUPT_ENABLE
#define E_DISABLE_BUS_INTERRUPT         SBA200E_INTERRUPT_DISABLE

#else /* ! SDC_ASIC */
#ifdef EBI_ASIC

#define LBAR_SHIFT			20
#define EISA200E_LBAR_ADDR_MASK		0x3f
#define EISA200E_LBAR_INTR_TYPE_LEVEL   0x00	/* should be ORed with the addr */  
#define EISA200E_LBAR_INTR_TYPE_TRIGGER 0x40    /* should be ORed with the addr */
#define	EISA200E_LBAR_INTR_ENABLE	0x00	/* should be ORed with the addr */
#define	EISA200E_LBAR_INTR_DISABLE	0x80	/* should be ORed with the addr */
#define HBAR_SHIFT			26
#define EISA200E_HBAR_ADDR_MASK		0x3f
#define EISA200E_HBAR_INTR_SEL_5	0x00	/* should be ORed with the addr */
#define EISA200E_HBAR_INTR_SEL_10	0x40	/* should be ORed with the addr */
#define EISA200E_HBAR_INTR_SEL_11	0x80	/* should be ORed with the addr */
#define EISA200E_HBAR_INTR_SEL_12	0xc0	/* should be ORed with the addr */

#define HCR_MASK                       	0x03 

#define E_BOARD_RESET                   ESA200E_BOARD_RESET_SET
#define E_CLEAR_BOARD_RESET             ESA200E_BOARD_RESET_CLEAR
#define E_HOLD_960_BUS                  ESA200E_HOLD_960_BUS
#define E_CLEAR_HOLD_960_BUS            ESA200E_RELEASE_960_BUS
#define E_INTERRUPT_960                 ESA200E_INTERRUPT_960A
#define E_CLEAR_INTERRUPT               ESA200E_CLEAR_INTERRUPT

#else /* ! EBI_ASIC */

#define HCR_MASK                        0x0

#define E_BOARD_RESET                   0
#define E_CLEAR_BOARD_RESET             0
#define E_HOLD_960_BUS                  0
#define E_CLEAR_HOLD_960_BUS            0
#define E_INTERRUPT_960                 0
#define E_CLEAR_INTERRUPT               0
#define E_ENABLE_BUS_INTERRUPT          0
#define E_DISABLE_BUS_INTERRUPT         0

#endif /* EBI_ASIC */
#endif /* SDC_ASIC */

/*********              END     BUS   ASIC  RELATED DEFINITIONS   **********/

#ifdef VME
#define BCR_OFFSET	0x400
#define VECREG_OFFSET	0x800
#endif

#ifdef EISA
/*
 *  PROM offset for various adapter data.
 */
#define BCR_OFFSET                 0
#define HBAR_OFFSET		   0x400
#define LBAR_OFFSET		   0x404
#define EISA_ID_PROM_OFFSET        0xc80
#define HW_TYPE_PROM_OFFSET        0xc84
#define HW_VERSION_PROM_OFFSET     0xc88
#define FW_VERSION_PROM_OFFSET     0xc8c
#define SW_VERSION_PROM_OFFSET     0xc90
#define SERIAL_NUM_PROM_OFFSET     0xc94
#define TX_FIFO_SIZE_PROM_OFFSET   0xc98
#define RX_FIFO_SIZE_PROM_OFFSET   0xc9c
#define MAC_ADDR_PROM_OFFSET       0xca0
#endif /* EISA */

#ifdef GIO
#define GIA100_ID 0x7c
#define GIA200_ID 0x60
#define PIW_REG_OFFSET  0x0
#define GIO_PROM_OFFSET 0x080000
#define BCR_OFFSET      0x0c0000
#define GIO_RAM_OFFSET  0x100000
#endif /* GIO */

/*
 *  Media types of network module daughter board
 *
 *  NOTE: The addition of media types must also be done in our SNMP
 *        (snmp/agent/fa_vars.h and snmp/agent/fa_vars.c).  The 
 *        fore-adapter.mib must also change, contact a member of
 *        the Network Management group for this.
 */
#define MT_MMODE_100		0x0	/* Multimode ODL at 100 Mbps */
#define MT_MMODE_140		0x1	/* Multimode ODL at 140 Mbps */
#define MT_MMODE_OC3_ST		0x2	/* Multimode SONET at 155 Mbps ST */
#define MT_MMODE_OC3_SC		0x3	/* Multimode SONET at 155 Mbps SC */
#define MT_SMODE_100		0x10	/* Singlemode ODL at 100 Mbps */
#define MT_SMODE_140		0x11	/* Singlemode ODL at 140 Mbps */
#define MT_SMODE_OC3_SHORT	0x12	/* 155 Mbps Singlemode short range */
#define MT_SMODE_OC3_LONG	0x13	/* 155 Mbps Singlemode long range */
#define MT_CAT5_UTP		0x20	/* 155 Mbps Category 5 UTP */
#define MT_UTP5_SONET           0x20    /* Unshielded twisted pair, 155 Mbps SONET */
#define MT_STP1_SONET           0x21    /* Shielded twisted pair, 155 Mbps SONET */
#define MT_ASX_100		0xF0	/* ASX-100 switch */
#define MT_ASX_200		0xF1	/* ASX-200 switch */

        /* the following are also defined in esp/include/esp/esp_baord.h */
#define MT_ESP_TAXI_140         0x4     /* ESP based taxi 140 */
#define MT_ESP_TAXI_100         0x5
#define MT_ESP_OC3              0x6


/*
 * Board Control Register. Write only.
 */
#define SBA200_BOARD_RESET        0x00000001     /* reset the i960 */
#define SBA200_CLEAR_BOARD_RESET  0x00000000
#define SBA200_CLEAR_INTERRUPT    0x00000004     /* clear interrupt issued by i960 */
#define SBA200_INTERRUPT_960      0x00000003     /* interrupt the i960 */
#define SBA200_CLEAR_HOLD_960_BUS 0x00000000
#define SBA200_HOLD_960_BUS       0x00000002     /* lock i960 from board RAM access */

/*
 * SBA300 Board Control Register. Write only.
 */
#define SBA300_BOARD_RESET        0x00000001     /* reset the i960 */
#define SBA300_CLEAR_BOARD_RESET  0x00000000
#define SBA300_HOLD_960_BUS       0x00000002     /* lock i960 from board RAM access */
#define SBA300_CLEAR_HOLD_960_BUS 0x00000000
#define SBA300_INTERRUPT_960      0x00000004     /* interrupt the i960 */
#define SBA300_CLEAR_INTERRUPT    0x00000008     /* clear interrupt issued by i960 */
#define SBA300_ESP_RESET          0x00000010     /* reset the ESP chip on the SBA300 */
#define SBA300_CLEAR_ESP_RESET    0x00000000     /* remove the reset of the ESP chip  */

/*
 *  EISA
 */
#define EISA_BOARD_RESET        0
#define EISA_CLEAR_BOARD_RESET  1
#define EISA_HOLD_960_BUS       2
#define EISA_CLEAR_HOLD_960_BUS 3
#define EISA_INTERRUPT_960      4
#define EISA_CLEAR_INTERRUPT    5
#define EISA_USE_IRQ5           6
#define EISA_USE_IRQ11          7
#define EISA_USE_IRQ10          8
#define EISA_USE_IRQ12          9

/* VME: set all bits except the last two, which are the only ones
** that are decoded on the board.
*/
#define VMA200_BOARD_RESET        0xFFFFFFFD     /* (1) reset the i960 */
#define VMA200_CLEAR_BOARD_RESET  0xFFFFFFFC
#define VMA200_HOLD_960_BUS       0xFFFFFFFE     /* (2) lock i960 from board RAM access */
#define VMA200_CLEAR_HOLD_960_BUS 0xFFFFFFFC
#define VMA200_INTERRUPT_960      0xFFFFFFFF     /* (3) interrupt the i960 */

#ifdef SBA300_BCR

/*
 * The ESP Timer is a fixed 24 bit timer with an 8 bit prescalar that allows the timer
 * increment to correspond to a cell time. The prescalar is driven from PCLK. 
 * If the prescalar is set properly the timer rolls over every 2^24 (12777216) 
 * cell times.
 */
#define	PCLK_FREQ	25000000.0
#define	PCLK_PERIOD	(1.0/PCLK_FREQ)
#define TAXI_100_CELL_TIME	((1.0/100000000) * 55 * 8)
#define TAXI_140_CELL_TIME	((1.0/140000000) * 55 * 8)
#define SONET_OC3_CELL_TIME	((1.0/149700000) * 53 * 8)
#define MAX_CELL_RATE_COUNT     65536.0 /* 2^16 */
/* 
 * SBA-300 Board Control Register defines
 */
#define BOARD_RESET		SBA300_BOARD_RESET
#define CLEAR_BOARD_RESET	SBA300_CLEAR_BOARD_RESET
#define HOLD_960_BUS		SBA300_HOLD_960_BUS
#define CLEAR_HOLD_960_BUS	SBA300_CLEAR_HOLD_960_BUS
#define INTERRUPT_960		SBA300_INTERRUPT_960
#define CLEAR_INTERRUPT		SBA300_CLEAR_INTERRUPT
#define ESP_BCR_RESET		SBA300_ESP_RESET
#define CLEAR_ESP_BCR_RESET	SBA300_CLEAR_ESP_RESET

#else /* ! SBA300_BCR = SBA200 */

#ifdef SBA200_BCR

#ifdef VME
#define BOARD_RESET		VMA200_BOARD_RESET
#define CLEAR_BOARD_RESET	VMA200_CLEAR_BOARD_RESET
#define HOLD_960_BUS		VMA200_HOLD_960_BUS
#define CLEAR_HOLD_960_BUS	VMA200_CLEAR_HOLD_960_BUS
#define INTERRUPT_960		VMA200_INTERRUPT_960
/* VME does not have a CLEAR_INTERRUPT function; it is done
** automatically by the hardware.
*/

#else /* VME */

#define BOARD_RESET		SBA200_BOARD_RESET
#define CLEAR_BOARD_RESET	SBA200_CLEAR_BOARD_RESET
#define HOLD_960_BUS		SBA200_HOLD_960_BUS
#define CLEAR_HOLD_960_BUS	SBA200_CLEAR_HOLD_960_BUS
#define INTERRUPT_960		SBA200_INTERRUPT_960
#define CLEAR_INTERRUPT		SBA200_CLEAR_INTERRUPT

#endif /* VME */

#else /* it's EISA */

#define BOARD_RESET		EISA_BOARD_RESET
#define CLEAR_BOARD_RESET	EISA_CLEAR_BOARD_RESET
#define HOLD_960_BUS		EISA_HOLD_960_BUS
#define CLEAR_HOLD_960_BUS	EISA_CLEAR_HOLD_960_BUS
#define INTERRUPT_960		EISA_INTERRUPT_960
#define CLEAR_INTERRUPT		EISA_CLEAR_INTERRUPT
#define USE_IRQ5		EISA_USE_IRQ5
#define USE_IRQ11		EISA_USE_IRQ11
#define USE_IRQ10		EISA_USE_IRQ10
#define USE_IRQ12		EISA_USE_IRQ12

#endif /* SBA200_BCR */
#endif /* SBA300_BCR */
/*      macros for handling bus ASICs   */

#define RESET_BOARD(au) {       			\
        if( au->au_unit_type == SERIES_200 ) {          \
            *au->au_bcr = BOARD_RESET ;                 \
        } else if( au->au_unit_type == SERIES_200E ) {  \
           volatile u_int hcr = *au->au_bcr ; /* Read hcr/bcr */        \
           *au->au_bcr = (hcr & HCR_MASK) | E_BOARD_RESET ;     \
        }       \
     }

#define CLEAR_RESET_BOARD(au) {                         \
        if( au->au_unit_type == SERIES_200 )    {       \
            *au->au_bcr = CLEAR_BOARD_RESET ;           \
        } else if( au->au_unit_type == SERIES_200E ) {  \
           volatile u_int hcr = *au->au_bcr ; /* Read hcr/bcr */        \
           *au->au_bcr = (hcr & HCR_MASK) & E_CLEAR_BOARD_RESET ;       \
        }       \
     }

#define CLEAR_BUS_INTERRUPT(au) {                       \
        if( au->au_unit_type == SERIES_200 ) {          \
           *au->au_bcr = CLEAR_INTERRUPT ;              \
        } else if( au->au_unit_type == SERIES_200E ) {  \
           volatile u_int hcr = *au->au_bcr ; /* Read hcr/bcr */        \
           *au->au_bcr = (hcr&HCR_MASK) | E_CLEAR_INTERRUPT ;           \
        }       \
    }

#define ENABLE_BUS_INTERRUPT(au) {      /* only for SERIES 200E */      \
        if( au->au_unit_type == SERIES_200E ) {         \
           volatile u_int hcr = *au->au_bcr; /* Read hcr/bcr */ \
           *au->au_bcr = (hcr & HCR_MASK) | E_ENABLE_BUS_INTERRUPT;     \
        }       \
    }

#define DISABLE_BUS_INTERRUPT(au) {     /* only for SERIES 200E */      \
        if( au->au_unit_type == SERIES_200E ) {         \
           volatile u_int hcr = *au->au_bcr; /* Read hcr/bcr */ \
           *au->au_bcr = (hcr & HCR_MASK) & E_DISABLE_BUS_INTERRUPT;    \
        }       \
    }

#endif /* _board_h */
