/*
 * Copyright (c) 1993 by Fore Systems, Inc.
 */
#ifndef _ncomm_h
#define _ncomm_h

#ifndef lint
static char _ncomm_h_rcsid[] = "@(#)$Id: ncomm.h,v 1.9 1994/09/10 18:58:27 tt Exp $ FSI";
#endif /* lint */

/*
 * Intel distributes source code for Mon960 running on boards
 * with each of the i960 core architectures CA, KA, KB and MC.
 * Mon960 for the EV80960CA (Intel i960 evaluation board) is
 * a CA architecture example. Mon960 for the EV80960CA and host
 * agents communicate via a serial line.
 *
 * Mon960 running on Fore Systems' 200-series line of adaptors 
 * and host agents communicate through shared memory. This file
 * defines the protocol: a simple character at a time with
 * a busy wait poll.
 *
 * comm960 <--{pseudo terminal}--> ncomm <---{200 RAM}---> Mon960
 *
 * NOTE: slave access to the 200-series is word access only.
 */
#define UART_OFFSET 0x00000400 /* offset from begining of RAM */
#define READY_XMIT	0x00000000 /* ready to transmit */
#define AVAIL	    0x01000000 /* character available */

struct soft_uart {
	u_int transmit_to_i960; /* and receive from host */
	u_int transmit_to_host; /* and receive from i960 */
};
typedef volatile struct soft_uart Soft_UART;


#define INIT_UART(uart) { \
	(uart)->transmit_to_i960 = READY_XMIT; \
	(uart)->transmit_to_host = READY_XMIT; \
}

/*
 * Host character I/O.
 */
#if !defined (__hpux) && !defined(sgi) && !defined(aix) && !defined(SunOS5)
#define PUTC_i960(uart, c) { \
	(uart)->transmit_to_i960 = (u_int)(AVAIL | (c)); \
}

#define GETC_i960(uart, c) { \
	u_int x = (uart)->transmit_to_host; \
	(c) = (u_char)(x & 0x000000ff); \
	(uart)->transmit_to_host = READY_XMIT; \
}

#define CAN_PUTC_i960(uart, not_used_for_directly_mapped_access) \
	((uart)->transmit_to_i960 == READY_XMIT) 

#define CAN_GETC_i960(uart)	\
	((uart)->transmit_to_host & AVAIL)
#else
/*
 *  When host makes a ioctl call to see if a character can be
 *  sent to 960, the character is already packed in the buffer
 *  associated with the ioctl. The charater is sent by the
 *  driver if it is OK to send.
 */
#define PUTC_i960(uart, c) 

/*
 *  When host receives a character from 960 in a call to ioctl.
 *  the driver also writes READY_XMIT (in the same ioctl) to indicate 
 *  the host is ready for the next character.
 */
#define GETC_i960(uart, c) { \
	u_int x = (uart)->transmit_to_host; \
	(c) = (u_char)(x & 0x000000ff); \
}

#define CAN_PUTC_i960(uart, c)	WRITE_UART(UART_OFFSET, c)
#define CAN_GETC_i960(uart)	READ_UART(UART_OFFSET)
#endif /* !__hpux && !sgi && !aix && !SunOS5 */



/*
 * i960 character I/O.
 */
#define PUTC_HOST(uart, c) { \
	(uart)->transmit_to_host = (u_int)(AVAIL | (c)); \
}

#define GETC_HOST(uart, c) { \
	(c) = (u_char)((uart)->transmit_to_i960 & 0xff); \
	(uart)->transmit_to_i960 = READY_XMIT; \
}

#define CAN_PUTC_HOST(uart) \
	((uart)->transmit_to_host == READY_XMIT) 

#define CAN_GETC_HOST(uart)	\
	((uart)->transmit_to_i960 & AVAIL)

extern Soft_UART softuart;
extern Soft_UART *uart;
#endif /* _ncomm_h */
