#include <dos.h>                        // FP_SEG() FP_OFF() getvect() setvect() enable() disable()
#include <conio.h>                      // getch() kbhit()
#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include <stddef.h>                     // offsetof()
#include <string.h>                     // strstr()
#include <mem.h>                        // memset()
#include <errno.h>                      // extern errno
#include <alloc.h>                      // coreleft()

#include "../dpmi32/phys.h"
#include "../dpmi32/intr.h"
#include "../pci/pci.h"
#include "../intel/timestmp.h"          // class PentiumClock
#include "../integrat/constant.h"       // CARD_CLOCK_HEADER CARD_CLOCK_DRIVER
#include CARD_CLOCK_HEADER              // class CardClock
#include "../intel/i8254-2.h"           // class HardwareTimer
#include "../aal5/atm.h"                // struct AtmHeader
#include CARD_DRIVER_HEADER             // class AtmCardRunning
#include "../intel/byteswap.h"          // ntohl()
#include "../intel/fpu.h"               // struct FpuControl, get_fpu_control() set_fpu_control()
#include "../wattcp/tcp.h"              // sock_XXX()
#include "../integrat/buffer.h"         // class BufferBlock, class BufferQueue
#include "../dpmi32/flat.h"             // align()
#include "../integrat/dump.h"
#include "../integrat/trace.h"          // struct tracerec
#include "../integrat/fs2flows.h"       // flows_init() flows_per_block() flows_cleanup()



// conditional compilation switches
//
#define ALLOW_TIMER_INTERRUPTS          FALSE       // FALSE to debug handler code
#define AFFECT_TIMER_INTERRUPTS         FALSE       // FALSE to allow BIOS and its piggybacks to get their timeslices
#define ONE_CARD_PER_TIMER_IRQ          TRUE
#define DEBUG_STACK_OVERFLOW            FALSE
#define PCI_INIT_STOPS_EARLY            FALSE
#define PCI_INIT_SETS_PCI_TIMING        FALSE       // causes lockup on SuperMicro P5STE (82439HX) with 200 MHz CPU when TRUE
#define PCI_INIT_SETS_PCI_MULTI_TIMING  TRUE        // causes cards to keep up better
#define PCI_INIT_SETS_PCI_CONTROL       FALSE       // causes lockup on ?ASUS? (82439HX) when TRUE
#define PCI_INIT_SETS_CACHING           FALSE       // causes lockup on SuperMicro P5STE when DPMIMEM=MAXMEM 65535 and caching already set to 512M in BIOS setup (CMOS)
#define CHECK_MEMORY_CONSISTENCY        TRUE        // no time wasted except on startup and shutdown
#define CHECK_MEMORY_WHILE_RUNNING      FALSE       // extra checks while >=1 card still running, affects RX free ring unavail count
#define CHECK_MEMORY_WHILE_LOOPING      FALSE       // extra checks inside main loop, big performance hit
#define CHECK_MEMORY_WHILE_QUERYING     FALSE       // extra checks when queried, ?moderate? performance hit
#define CHECK_MEMORY_BEFORE_PROCESSING  FALSE       // extra checks before calling flows_per_block()
#define CHECK_MEMORY_AFTER_PROCESSING   FALSE       // extra checks after  calling flows_per_block()
#define LOCK_PAGES                      FALSE
#define ALLOW_STATS_CALLS               TRUE        // main call to flows_per_block()
#define ALLOW_STATS_CALLS_2             TRUE        // main call to analyze_partial_blocks() when empty queue
#define ALLOW_STATS_CALLS_3             TRUE        // call to flows_query()
#define BUSYWAIT_STATS_CALLS            TRUE        // only if not allowed
#define BUSYWAIT_STATS_CALLS_2          FALSE       // not implemented, only if not allowed
#define BUSYWAIT_STATS_CALLS_3          FALSE       // only if not allowed
#define TIME_STATS_CALLS                TRUE        // main call to flows_per_block()
#define TIME_STATS_CALLS_2              FALSE       // not implemented, extra call to flows_per_block() when queried and queue is empty
#define TIME_STATS_CALLS_3              TRUE        // call to flows_query()
#define SIMULATE_FIBER                  FALSE       // when TRUE, please don't hook up any physical fiber
#define TEST_DIVISION                   TRUE
#define DO_FULLNESS_STATS               TRUE        // keep/display min/max/mean of BufferQueue filled list
#define DO_FREENESS_STATS               TRUE        // keep/display min/max/mean of BufferQueue free list
#define TRACE_BUFFER_BLOCKS             FALSE       // save a BufferBlock we predict will be bad to disk
#define CHECK_ATM_INTERRUPT_ERRORS      TRUE        // polls in main loop



// compile-time switches which depend on others
//
#define PCI_EXIT_CALLS_BUFFER_DUMP      !DO_FOREGROUND_STATS

#if CARD_IS_OC12 || (CARD_COPIES_ATM_PAYLOAD && !CARD_COPIES_ALL_CELLS)
#define CARD_HANDLES_TIMESTAMP_WRAP     TRUE
#else
#define CARD_HANDLES_TIMESTAMP_WRAP     FALSE
#endif



// manifest constants
//
//#define TIMER_FREQUENCY   65536       // should be every 55.07 milliseconds
//#define TIMER_FREQUENCY   12100       // is every 10.063 milliseconds
//#define TIMER_FREQUENCY   11900       // should be every 10.0 milliseconds
//#define TIMER_FREQUENCY   3840        // is every 3.21 milliseconds
//#define TIMER_FREQUENCY   3500        // is every 2.79 milliseconds
//#define TIMER_FREQUENCY   1279        // is every .8542 milliseconds
//#define TIMER_FREQUENCY   1280        // is every 1.066 milliseconds
//#define TIMER_FREQUENCY   1190        // should be every 1.00 milliseconds, is .858 mS
//#define TIMER_FREQUENCY   900         // is every .642 milliseconds (also)
//#define TIMER_FREQUENCY   800         // is every .642 milliseconds
//#define TIMER_FREQUENCY   700         // is every .429 milliseconds     ***
#define TIMER_FREQUENCY   500         // is every .214 milliseconds
//#define TIMER_FREQUENCY   300         // is every .208 microseconds
//#define TIMER_FREQUENCY   250         // is every .208 microseconds

#define TIMER_CALIBRATE     3           // record start timestamp here

#define FLOW_BLOCK_TIME     400         // how long should dummy flows_per_block() take? (microseconds)
//#define FLOW_QUERY_TIME   56          // how long should dummy flows_query() take? (milliseconds)
#define FLOW_QUERY_TIME     0           // how long should dummy flows_query() take? (milliseconds)

#define DESIRED_MIN_GRANT   2

//#define DESIRED_CACHE_CONTROL    0x01   // (default w/o cache card) L1 cache enabled, L2 cache not forcing miss, NA# signal enabled, SRAM is 0K pipe burst
#define DESIRED_CACHE_CONTROL    0xB1   // (default with cache card installed) L1 cache enabled, L2 cache not forcing miss, L2 stops at 64 MB, NA# signal enabled, SRAM is 512K dual bank pipe burst
//#define DESIRED_CACHE_CONTROL    0xB0   // L1 cache disabled, L2 cache disabled, NA# signal enabled, SRAM is 512K dual bank pipelined burst
//#define DESIRED_CACHE_CONTROL    0xB3   // L1 cache enabled, L2 cache is forcing miss, NA# signal enabled, SRAM is 512K dual bank pipelined burst

//#define DESIRED_HOST_CONTROL   0xE0   // CPU inactivity timer = 8 PCI clocks
#define DESIRED_HOST_CONTROL     0x80   // (default) CPU inactivity timer = 5 PCI clocks
//#define DESIRED_HOST_CONTROL   0x40   // CPU inactivity timer = 3 PCI clocks
//#define DESIRED_HOST_CONTROL   0x20   // CPU inactivity timer = 2 PCI clocks
//#define DESIRED_HOST_CONTROL   0x00   // CPU inactivity timer = 1 PCI clocks

#define DESIRED_HOST_LATENCY        0   // works well when card latency is also 0
//#define DESIRED_HOST_LATENCY      8   // works just as badly as 64, regardless of card latency
//#define DESIRED_HOST_LATENCY     64   // AMI Triton BIOS default

#define DESIRED_MULTI_LATENCY      32   // ???works??? when CPU latency=32, card latency=64



// error return codes
//
#define ERROR_PCI_BASE                  3000
#define ERROR_PCI_TOO_FEW_CARDS_FOUND   (ERROR_PCI_BASE + 0)        // unused now
#define ERROR_PCI_CANNOT_ALLOC          (ERROR_PCI_BASE + 1)



unsigned timer_interrupt_card_index = 0;
unsigned long timer_interrupt_count = 0;
const unsigned timer_irq_level = 0;
unsigned char timer_interrupt;
vfp timer_thunk;
Bit48 old_timer_handler;
PentiumClock timer_start_time;
PentiumClock timer_end_time;
HardwareTimer *timer = 0;
class AtmCardRunning *cards[ MAX_ATM_NICS ? MAX_ATM_NICS : 1 ];
Bit8 old_pci_latency_timer;
#if INTEL_82439HX
    Bit8 old_pci_multi_timer;
#endif
static union { PciControl old_pci_control; Bit8 old_pci_control_alias; };
static union { PciCacheControl old_pci_cache_control; Bit8 old_pci_cache_control_alias; };
#if DEBUG_STACK_OVERFLOW
    int stack_ptr_recorded = FALSE;
    Bit16 stack_seg_during_main_top = 0;
    Bit16 stack_seg_during_main_loop = 0;
    Bit16 stack_seg_during_global_timer = 0;
    Bit32 stack_ptr_during_main_top = 0;
    Bit32 stack_ptr_during_main_loop = 0;
    Bit32 stack_ptr_during_global_timer = 0;
#endif
#if LOCK_PAGES
    int atm_shift_dummy_before( void );
#endif
#if TRACE_BUFFER_BLOCKS
    Bit8 *block_trace = (Bit8 *) 0x1F6F9000;
#endif



// find the ATM NIC's on the PCI bus
// and initialize them
//
int pci_init( LoopbackSelection loop_type, int want_rx, int want_tx )
{
    unsigned count;
    Bit8 value;
    Bit16 handle;
    int result;

    printf( "array of pointers to AtmCard is at 0x%p\n", cards );

#if CHECK_MEMORY_CONSISTENCY
    result = all_buffers->check();
    if( result )
        return result;
#endif

    // check whether PCI BIOS is present
    //
    if( !pci_is_bios_present() )
    {
        printf( "Sorry, PCI BIOS is not present.\n" );
        exit( 1 );
    }

    handle = 0;
    old_pci_latency_timer = pci_get_master_latency_timer( handle );
    printf( "CPU master-latency-timer register was %d (%g uS)\n", old_pci_latency_timer, old_pci_latency_timer*0.25 );

#if INTEL_82439HX
    old_pci_multi_timer = pci_get_multi_transaction_timer( handle );
    printf( "CPU multi-transaction-timer register was %d (%g uS)\n", old_pci_multi_timer, old_pci_multi_timer*0.25 );
#endif

    old_pci_control_alias = pci_get_control( handle );
#if INTEL_82437FX
    printf( "CPU PCI control register was 0x%2.2X (bus concurrency %s,"
        " PCI streaming %s, CPU-to-PCI write bursting %s, peer concurrency %s,"
        " CPU inactivity timer %d)\n",
        old_pci_control_alias,
        old_pci_control.bus_concurrency_disable           ? "disabled" : "enabled",
        old_pci_control.pci_streaming_disable             ? "disabled" : "enabled",
        old_pci_control.cpu_to_pci_write_bursting_disable ? "disabled" : "enabled",
        old_pci_control.peer_concurrency_enable           ? "enabled"  : "disabled",
        old_pci_control.cpu_inactivity_timer + 1
    );
#else
#if INTEL_82439HX
    printf( "CPU PCI control register was 0x%2.2X (global TXC %s,"
        " SERR# output is %s, peer concurrency is %s, NA# %s,"
        " CPU shutdown cycle causes %s, ECC mode is %s, DRAM has %s)\n",
        old_pci_control_alias,
        old_pci_control.global_txc_enable                 ? "enabled"  : "disabled",
        old_pci_control.serr_pound_output_type            ? "actively driven" : "open drain for PCI",
        old_pci_control.peer_concurrency_enable           ? "enabled"  : "disabled",
        old_pci_control.dual_processor_na_pound_enable    ? "uses policies for dual-procesor PCI 2.1" : "uses 82437FX policies",
        old_pci_control.shutdown_to_port_92               ? "write 01h to port 92" : "shutdown cycle on PCI bus",
        old_pci_control.ecc_test_enable                   ? "test" : "normal",
        old_pci_control.dram_ecc_or_parity_select         ? "ECC" : "parity"
    );
#endif
#endif

    old_pci_cache_control_alias = pci_get_cache_control( handle );
    printf( "CPU cache control register was 0x%2.2X (first level cache %s,"
        " secondary cache miss/invalidate %sforced,"
#if INTEL_82439HX
        " L2 cache up to %s,"
#endif
        " NA# signal %s, SRAM type %d, secondary cache size %dK)\n",
        old_pci_cache_control_alias,
        old_pci_cache_control.first_level_cache_enable ? "enabled" : "disabled",
        old_pci_cache_control.secondary_cache_force_miss_or_invalidate ? "is " : "not ",
#if INTEL_82439HX
        old_pci_cache_control.extended_cachability_enable ? "512 MB" : "64 MB",
#endif
        old_pci_cache_control.na_pound_signal_disable ? "disabled" : "enabled",
        old_pci_cache_control.static_ram_type,
        old_pci_cache_control.secondary_cache_size * 256
    );

#if PCI_INIT_SETS_PCI_MULTI_TIMING
    pci_set_multi_transaction_timer( handle, DESIRED_MULTI_LATENCY );
    value = pci_get_multi_transaction_timer( handle );
    if( value != DESIRED_MULTI_LATENCY )
        printf( "unable to set CPU multi-transaction-timer register to %d; is now %d (%g uS)\n", DESIRED_MULTI_LATENCY, value, value*0.25 );
    else
        printf( "changed CPU multi-transaction-timer register to %d (%g uS)\n", DESIRED_MULTI_LATENCY, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_TIMING
    pci_set_master_latency_timer( handle, DESIRED_HOST_LATENCY );
    value = pci_get_master_latency_timer( handle );
    if( value != DESIRED_HOST_LATENCY )
        printf( "unable to set CPU master-latency-timer register to %d; is now %d (%g uS)\n", DESIRED_HOST_LATENCY, value, value*0.25 );
    else
        printf( "changed CPU master-latency-timer register to %d (%g uS)\n", DESIRED_HOST_LATENCY, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_CONTROL
    pci_set_control( handle, DESIRED_HOST_CONTROL );
    value = pci_get_control( handle );
    if( value != DESIRED_HOST_CONTROL )
        printf( "unable to set CPU control register to 0x%2.2X; is now 0x%2.2X\n", DESIRED_HOST_CONTROL, value );
    else
        printf( "changed CPU control register to 0x%2.2X\n", value );
#endif

#if PCI_INIT_SETS_CACHING
    pci_set_cache_control( handle, DESIRED_CACHE_CONTROL );
    value = pci_get_cache_control( handle );
    if( value != DESIRED_CACHE_CONTROL )
        printf( "unable to set CPU cache control register to 0x%2.2X; is now 0x%2.2X\n", DESIRED_CACHE_CONTROL, value );
    else
        printf( "changed CPU cache control register to 0x%2.2X\n", value );
#endif

    // for all the ATM cards that we want...
    //
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        cards[ count ] = new AtmCardRunning( count + ATM_NIC_OFFSET );
        if( !cards[count] )
            return ERROR_PCI_CANNOT_ALLOC;
        if( cards[count]->init_result )
            return cards[count]->init_result;
    }
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        result = cards[count]->card_init( AtmCardRunning::NormalAtm, loop_type, want_rx, want_tx );
        if( result )
            return result;
    }

    printf( "PCI configuration discovery done\n\n" );

    return 0;
}



// shut down whatever cards we were using
//
int pci_exit( void )
{
    int count;
    Bit32 block_count;
    Bit32 buffer_count;
    int result;
    AtmCardRunning *this_card;

    // separate global messages from card messages
    //
    printf( "\n" );

    for( count=MAX_ATM_NICS-1; count>=0; count-- )
    {
        // we must set global pointer to nil before we
        // deallocate the card, because otherwise the next
        // timer interrupt will try to use the object after
        // it has been deallocated and possibly reused,
        // which is unsafe for MANY reasons
        //
        this_card = cards[ count ];
        cards[ count ] = NULL;
        delete this_card;

        // separate messages for one card from the other
        //
        printf( "\n" );
    }

    union { PciStatus status; Bit8 status_alias; };
    status_alias = pci_get_status( 0 );
    printf( "host PCI final status 0x%4.4X\n", status_alias );

#if CHECK_MEMORY_CONSISTENCY
    // don't want to exit when this fails because the
    // diagnostics that follow may still be useful
    //
    all_buffers->check();
#endif

    printf( "BufferQueue::allocate() failed %d times\n", all_buffers->allocate_failure_count );
    for( count=0; count<MAX_ATM_NICS; count++ )
        printf( "BufferQueue::read_tx(%d) failed %d times\n", count, all_buffers->read_tx_failure_count[count] );

    // dump the host buffer queue states
    //
    printf( "global host buffer queue has %ld free buffers\n", all_buffers->amount_free() );
    buffer_count = all_buffers->amount_filled( block_count );
    printf( "global host buffer queue has %ld/%ld filled buffers/blocks\n", buffer_count, block_count );
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        buffer_count = all_buffers->amount_sendable( count, block_count );
        printf( "global host buffer queue has %ld/%ld sendable buffers/blocks for cardindex %d\n", buffer_count, block_count, count );
    }

#if PCI_EXIT_CALLS_BUFFER_DUMP
    // dump the data in the global queue to disk
    //
    char filename[ 13 ];

    sprintf( filename, "D:\\IP.DMP" );
    printf( "pci_exit() saving IP packets with headers in file '%s'\n", filename );
    result = all_buffers->dump( filename );
    if( result != 0 )
    {
        printf( "error %d dumping queue\n", result );
    }
#endif

#if PCI_INIT_SETS_PCI_MULTI_TIMING || PCI_INIT_SETS_PCI_TIMING || PCI_INIT_SETS_PCI_CONTROL || PCI_INIT_SETS_CACHING
    Bit16 handle = 0;
    Bit8 value;
#endif

#if PCI_INIT_SETS_PCI_MULTI_TIMING
    // reset the global multi transaction timer
    //
    pci_set_multi_transaction_timer( handle, old_pci_multi_timer );
    value = pci_get_multi_transaction_timer( handle );
    if( value != old_pci_multi_timer )
        printf( "unable to restore CPU multi-transaction-timer register to %d; is now %d (%g uS)\n", old_pci_multi_timer, value, value*0.25 );
    else
        printf( "restored CPU multi-transaction-timer register to %d (%g uS)\n", old_pci_multi_timer, value*0.25 );
#endif

#if PCI_INIT_SETS_PCI_CONTROL
    // reset the control register for host
    //
    pci_set_control( handle, old_pci_control_alias );
    value = pci_get_control( handle );
    if( value != old_pci_control_alias )
        printf( "unable to restore CPU control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_control_alias, value );
    else
        printf( "restored CPU control register to 0x%2.2X\n", old_pci_control_alias );
#endif

#if PCI_INIT_SETS_PCI_TIMING
    // reset the master latency timer for host
    //
    pci_set_master_latency_timer( handle, old_pci_latency_timer );
    value = pci_get_master_latency_timer( handle );
    if( value != old_pci_latency_timer )
        printf( "unable to restore CPU master-latency-timer register to %d; is now %d (%g uS)\n", old_pci_latency_timer, value, value*0.25 );
    else
        printf( "restored CPU master-latency-timer register to %d (%g uS)\n", old_pci_latency_timer, value*0.25 );

    // reset the control register for host
    //
    pci_set_control( handle, old_pci_control_alias );
    value = pci_get_control( handle );
    if( value != old_pci_control_alias )
        printf( "unable to restore CPU control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_control_alias, value );
    else
        printf( "restored CPU control register to 0x%2.2X\n", old_pci_control_alias );
#endif

#if PCI_INIT_SETS_CACHING
    // reset the cache control register for host
    //
    // note: it is very important not to restore the caching
    // settings until we are done messing with high memory
    // (i.e. dumping it to disk) because we might lock up
    // the machine otherwise!
    //
    pci_set_cache_control( handle, old_pci_cache_control_alias );
    value = pci_get_cache_control( handle );
    if( value != old_pci_cache_control_alias )
        printf( "unable to restore CPU cache control register to 0x%2.2X; is now 0x%2.2X\n", old_pci_cache_control_alias, value );
    else
        printf( "restored CPU cache control register to 0x%2.2X\n", old_pci_cache_control_alias );
#endif

    // delete of all_buffers should be here
    //

    return result;
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
// P.S. I have no clue why I don't have to do this
// for class static functions AtmCard::handler1()
// and AtmCard::handler2()
//
#pragma warn -par

// deals with interrupts from hardware timer chip
// on Pentium PC motherboard
//
// should occur at least often enough to poll the SAR
// so the receive buffers do not overflow
//
int timer_handler(
    void *parms,
    Bit32 ebp,
    Bit32 edi,
    Bit32 esi,
    Bit32 ds,
    Bit32 es,
    Bit32 edx,
    Bit32 ecx,
    Bit32 ebx,
    Bit32 eax,
    Bit32 eip,
    Bit32 cs,
    Bit32 flags
)
{
    // count the number of interrupts for debugging
    //
    timer_interrupt_count++;

    // I need help calibrating the pentium timestamps,
    // so record 2 timestamps, one interrupt apart
    //
    if( timer_interrupt_count == TIMER_CALIBRATE )
        timer_start_time.set();
    else if( timer_interrupt_count == TIMER_CALIBRATE + 1 )
        timer_end_time.set();

#if DEBUG_STACK_OVERFLOW
    if( !stack_ptr_recorded )
    {
        stack_ptr_during_global_timer = _ESP;
        stack_seg_during_global_timer = _SS;
    }
#endif

    // any polling of SAR chips should occur here
    //
#if ONE_CARD_PER_TIMER_IRQ
    // note that we only call one card per interrupt
    //
    if( cards[ timer_interrupt_card_index ] )
        cards[ timer_interrupt_card_index ]->timer_handler();

    timer_interrupt_card_index++;
    if( timer_interrupt_card_index >= MAX_ATM_NICS )
        timer_interrupt_card_index = 0;
#else
    if( cards[0] )
        cards[0]->timer_handler();

#if MAX_ATM_NICS >= 2
    if( cards[1] )
        cards[1]->timer_handler();
#endif
#endif

    // tell the interrupt controllers we are done
    //
    AtmCardRunning::irq.acknowledge( timer_irq_level );

    // tell the thunking code not to chain to the old handler
    //
    return 0;
}

// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par




void timer_init( void )
{
    // calculate the software interrupt numbers
    // that the CPU will see for the IRQ levels
    //
    timer_interrupt = irq_to_intr( timer_irq_level );

    // disable maskable interrupts
    //
    disable();

    // save the old interrupt vectors
    //
    old_timer_handler = getvect( timer_interrupt );

#if AFFECT_TIMER_INTERRUPTS
    // create a thunk so our C++ function can be called at interrupt time
    //
    timer_thunk = make_thunk( (ifp)timer_handler, &old_timer_handler, NULL );
#if THROW_IS_LEGAL
    if( timer_thunk == NULL )
        throw "error!  cannot allocate timer interrupt thunk!\n";
#endif

    // point the vector at our handler
    //
    Bit48 pointer = { (Bit32)timer_thunk, _CS };
    setvect( timer_interrupt, pointer );

    // allow desired IRQ level through to the processor
    //
#if ALLOW_TIMER_INTERRUPTS
    AtmCardRunning::irq.unmask( timer_irq_level );
#else
    AtmCardRunning::irq.mask( timer_irq_level );
#endif

    // initialize hardware timer 0
    //
    timer = new HardwareTimer;

    // set the value that hardware timer 0
    // will decrement until interrupt
    //
    timer->set( TIMER_FREQUENCY );
#endif

    // enable maskable interrupts
    //
    enable();
}



// restore interrupt handling to the way it was when
// we started up
//
void timer_exit( void )
{
    // disable maskable interrupts
    //
    disable();

#if AFFECT_TIMER_INTERRUPTS
    // reset the clock frequency to the one we think BIOS likes
    //
    timer->set( 65536 );

    // stop using hardware timer 0
    //
    delete timer;

    // point the vectors at the old handlers
    //
    // must be before we delete irq or we will
    // see an extraneous timer interrupt
    //
    // I know, I know, interrupts are supposed
    // to be disabled, but I'm not lying!
    //
    setvect( timer_interrupt, old_timer_handler );

    // free up the 32-bit assembly thunking code
    //
    free( timer_thunk );
#endif

    // enable maskable interrupts
    //
    enable();

    // print out what meager statistics we
    // have for the hardware timer chip
    //
    printf( "hardware timer 0 caused %ld interrupts\n", timer_interrupt_count );
    double elapsed = timer_end_time - timer_start_time;
    printf( "%g pentium clocks elapsed between hardware timer 0 interrupts %d and %d\n",
        elapsed, TIMER_CALIBRATE, TIMER_CALIBRATE+1 );
}



#if DO_FOREGROUND_STATS
// do stats analysis on the partially-received blocks
//
void analyze_partial_blocks( void )
{
    unsigned count;
    int result;

    for( count=0; count<MAX_ATM_NICS; count++ )

        // beware of NULL block pointer in AtmCard
        // which indicates that block allocation
        // failed last time around
        //
        if( cards[count]->my_block )
        {
            cards[count]->set_pending_block_usage();
            result = flows_per_block( cards[count]->my_block );
            if( result )
                fprintf( stderr, "error %d from partial flows_per_block()!\n", result );
        }
}
#endif



int main( int argc, char *argv[] )
{
    int keystroke;
    int result;
    int read_result;
    Bit32 count;
    Bit8 interface;
#if DO_FOREGROUND_STATS
    BufferBlock *block;
    unsigned block_count = 0;
    Bit64 buffer_count = (Bit32) 0;
    PentiumClock before, after;
    PentiumClock startup_time;
    PentiumClock active_interval, active_alarm;
    PentiumClock query_interval, query_alarm;
    double elapsed2;
    char query_msg[ 2000 ] = "";
    int query_msg_len = 0;
    Bit32 queried_allocate_failures = 0;
    Bit32 queried_read_tx_failures[ MAX_ATM_NICS ];
#if TIME_STATS_CALLS
    PentiumClock elapsed, elapsed_min, elapsed_max, elapsed_tot;
    double elapsed_min2, elapsed_max2, elapsed_avg2;
    elapsed_min.high = -1;
    elapsed_min.low  = -1;
    elapsed_max.high = 0;
    elapsed_max.low  = 0;
    elapsed_tot.high = 0;
    elapsed_tot.low  = 0;
#endif
#if DO_FULLNESS_STATS
    Bit32 max_filled, min_filled, count_filled;
    Bit64 sum_filled;
#endif
#if DO_FREENESS_STATS
    Bit32 max_free, min_free, count_free;
    Bit64 sum_free;
#endif
#if !ALLOW_STATS_CALLS && BUSYWAIT_STATS_CALLS
    Bit32 busy_iterations = 0;
#endif
#endif
    PentiumClock now;
    int want_facility_loopback;
    int want_rx;
    int want_tx;
    PentiumClock keystroke_interval, keystroke_alarm;
    FpuControl float_control;
    union { FpuStatus float_status; Bit16 float_status_alias; };
    Bit8 *data;



    float_control = get_fpu_control();

    // force floating point division by zero to generate
    // negative or positive infinity, depending on the
    // sign of the numerator, instead of interrupt 0x10
    //
    // this is the same way DEC Alpha with GNU C behaves
    //
    float_control.zero_divide_mask = TRUE;

    // force conversion of a floating point infinity to
    // integer to succeed, instead of ?interrupt 0x10
    //
    // this is the behavior of DELL and Aberdeen PC's
    // since they have a null handler for that exception
    //
    float_control.invalid_operation_mask = TRUE;

    set_fpu_control( float_control );

#if TEST_DIVISION
    // try dividing a 32 bit unsigned int by float zero
    //
    Bit32 divtest_numerator = 144765;
    double divtest_denominator = 0.0;
    double divtest_result = divtest_numerator / divtest_denominator;
    printf( "Bit32 %u / double %f = double %f\n",
        divtest_numerator,
        divtest_denominator,
        divtest_result
    );

    // now try converting the result to an integer
    //
    Bit32 divtest_result_integer = divtest_result;
    printf( "double %f converts to Bit32 %u\n", divtest_result, divtest_result_integer );

    // now try dividing float zero by float zero
    //
    double divtest_numerator_float = 0.0;
    divtest_result = divtest_numerator_float / divtest_denominator;
    printf( "double %f / double %f = double %f\n",
        divtest_numerator_float,
        divtest_denominator,
        divtest_result
    );

    // now try converting the new result to an integer
    //
    divtest_result_integer = divtest_result;
    printf( "double %f converts to Bit32 %u\n", divtest_result, divtest_result_integer );

#if 0       // don't bother, it always fails
    // now try dividing a 32 bit unsigned int by an integer zero
    //
    Bit32 divtest_denominator_integer = 0;
    divtest_result_integer = divtest_numerator / divtest_denominator_integer;
    printf( "Bit32 %u / Bit32 %u = Bit32 %u\n",
        divtest_numerator,
        divtest_denominator_integer,
        divtest_result_integer
    );
#endif

    // now clear out all the floating point exceptions
    // so main() doesn't complain about them
    //
    clear_fpu_status();
#endif



    // init TCP/IP library
    //
    // we must do this before any other physical memory
    // function calls because those must be undone in the
    // same order that they were done, and sock_init()
    // installs an atexit() handler which means we are not
    // in control of when sock_exit() gets called
    //
    sock_init();



    // initialize access to our own area of flat memory
    // which we know how to convert to physical addresses
    //
    // note that DOS 6.22 HIMEM.SYS cannot access more than
    // 64 MB of physical memory, although WIN 95 HIMEM.SYS
    // can do at least 128 MB
    //
    // note that DPMI32VM.OVL included with Borland PowerPack
    // 1.0 for DOS truncates the value of DPMIMEM=MAXMEM 70000
    // to 16 bits, and hence will not try to allocate more than
    // 64 MB minus 1 KB of memory, although you can patch file
    // offset 35E0h from 2B 1E 66 06 to 90 90 90 90 so it will
    // allocate until memory is exhausted (ignoring environment
    // variables)
    //
    flat_init();

#if LOCK_PAGES
    // try to lock the whole program
    //
    Bit16 lock_result = lock_linear( flat_to_linear(&atm_shift_dummy_before), 0x3000 );
    printf( "got error 0x%4.4X locking pages of 0x%8.8lX AtmCard::atm_shift_hw_rings()\n", lock_result, atm_shift_dummy_before );
#endif

#if DEBUG_STACK_OVERFLOW
    // say what's up with the stack
    //
    printf( "stack is at log/lin 0x%4.4X:%8.8lX 0x%8.8lX\n",
        _SS, _ESP, flat_to_linear((void *)_ESP) );
    printf( "stack segment has base 0x%8.8lX and limit 0x%8.8lX\n",
        get_base(_SS), get_limit(_SS) );

    if( !stack_ptr_recorded )
    {
        stack_ptr_during_main_top = _ESP;
        stack_seg_during_main_top = _SS;
    }
#endif


    // no BufferQueue::read_tx() failures have been queried yet
    //
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        queried_read_tx_failures[ interface ] = 0;

    // default card initialization to doing facility loopback
    //
    want_facility_loopback = TRUE;

    // default card to do RX but not TX
    //
    want_rx = TRUE;
    want_tx = FALSE;

    // keyboard will still be checked when foreground stats
    // are not being done, so default its alarm to once a second
    //
    keystroke_interval.put_seconds( 1.0 );

    // just get the time once for the many interval calculations
    //
    now.set();

#if DO_FOREGROUND_STATS
    // in case flows module doesn't fill in the variables, also
    // default query and active feed intervals to never
    //
    query_interval.high = -1;
    query_interval.low = -1;
    query_interval -= now;
    active_interval.high = -1;
    active_interval.low = -1;
    active_interval -= now;

    // tell the stats code about the command-line arguments
    // and give it a chance to alter the timeouts
    //
    fprintf( stderr, "%u bytes of RAM initially available to malloc()\n", coreleft() );
    fprintf( stderr, "%u bytes of RAM initially available to flat_alloc()\n", flat_remaining() );
    result = flows_init( argc, argv, &keystroke_interval, &query_interval, &active_interval, &want_facility_loopback, &want_rx, &want_tx );
    fprintf( stderr, "%u bytes of RAM are still available to malloc()\n", coreleft() );
    fprintf( stderr, "%u bytes of RAM are still available to flat_alloc()\n", flat_remaining() );
    if( result )
        return result;

    // now set up initial values for all the alarms
    //
    query_alarm = now + query_interval;
    active_alarm = now + active_interval;
#endif
    keystroke_alarm = now + keystroke_interval;

    // allocate all remaining flat memory to card buffer pool
    //
    // we must do this AFTER statistics initialize, so they
    // get first crack at flat memory, in case they are using
    // it instead of malloc()
    //
    all_buffers = new BufferQueue;
    if( all_buffers == NULL )
        goto cleanup;

    // say where the global buffers are
    //
    printf( "global control blocks are at 0x%p\n", &all_buffers );

    data = all_buffers->get_start()->buffers;
    printf( "global data buffers are at   log/lin/phy"
        " 0x%4.4X:%8.8lX 0x%8.8lX 0x%8.8lX\n", _DS, data,
        flat_to_linear(data), flat_to_physical(data) );

    printf( "SW_RINGS=%d\n", SW_RINGS );

    printf( "COMPL_RING_ENTRIES=%d\n", COMPL_RING_ENTRIES );

#if DO_FOREGROUND_STATS && DO_FULLNESS_STATS
    // initialize BufferQueue fullness statistics
    //
    max_filled = 0;
    min_filled = BIT32_MAX;
    sum_filled = Bit32( 0 );
    count_filled = 0;
#endif

#if DO_FOREGROUND_STATS && DO_FREENESS_STATS
    // initialize BufferQueue fullness statistics
    //
    max_free = 0;
    min_free = BIT32_MAX;
    sum_free = Bit32( 0 );
    count_free = 0;
#endif

    // initialize the h/w timer and interrupt controller chips
    //
    timer_init();

    // get the addresses and interrupts of
    // the PCI ATM cards
    //
    result = pci_init( want_facility_loopback ? Facility : NoLoopback, want_rx, want_tx );
    if( result )
    {
        printf( "error %d from pci_init()\n", result );
        goto cleanup;
    }

    // synchronize the clocks on the cards by issuing the
    // "clear your timestamp" command to them all in rapid
    // sequence, then checking results after
    //
    // assumes all card pointers in array have been filled
    // in, which is guaranteed by pci_init()
    //
    // assumes functions called do their own error messages
    //
    CardClock::sync_time.set();
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        result = cards[ count ]->start_clearing_timestamp();
        if( result )
            goto cleanup;
    }
    for( count=0; count<MAX_ATM_NICS; count++ )
    {
        result = cards[ count ]->wait_for_timestamp_to_clear();
        if( result )
            goto cleanup;
    }

#if !CARD_INIT_PROVIDES_BUFFERS
    if( want_rx )

        // give one BufferBlock to each card,
        // to allow them to finally start receive
        //
        // we avoided it until now so init would work
        // even if the link is very busy
        //
        for( count=0; count<MAX_ATM_NICS; count++ )
        {
            cards[ count ]->foreground_replenish();
        }

    // for transmit, whoever calls BufferQueue::fill_tx()
    // has the job of calling that card's foreground_tx_replenish()
    //
#endif

    // note: we cannot flush the level 1 cache because
    // the Borland PowerPack for DOS 32-bit protected
    // mode environment puts us in ring 3, so executing
    // the WBINVD instruction would cause a protection
    // violation
    //
    // we cannot fill the code cache either, because the
    // Pentium processor maintains separate data and code
    // caches, so reads of the code segment would only
    // affect the data cache
    //

    printf( "\n" );


#if DO_FOREGROUND_STATS && SIMULATE_FIBER
    // create a dummy block that has all invalid buffers
    //
    result = all_buffers->allocate( &block );
    if( result )
    {
        printf( "fatal error: main() has error %d allocating a dummy block\n", result );
        return result;
    }

    block->cardindex = 0;
    block->cardnum = cards[ 0 ]->pci_handle;
    block->used = COMPL_RING_ENTRIES;
    {
        PentiumClock now;
        now.set();
        CardClock now2 = now;
        block->timestamp() = now2;
    }

    for( count=0; count<COMPL_RING_ENTRIES; count++ )
    {
        // fudge a nonzero IP fragment offset offset
        // so flows_per_block() will ignore the data
        //
        *(block->buffers + count*BUFFER_SIZE + 12 + 11) = 1;
    }
#endif

#if DO_FOREGROUND_STATS
    // remember when we started collecting stats
    //
    startup_time.set();
#endif

    // wait for the user to unplug fiber(s)
    // to stop SAR from filling its memory
    // while we empty it
    //
    printf( "\nplease reconnect fibers\npress escape when you want to stop capturing\n" );
    while( TRUE )
    {
#if DEBUG_STACK_OVERFLOW
        stack_ptr_during_main_loop = _ESP;
        stack_seg_during_main_loop = _SS;
#endif

#if CHECK_MEMORY_WHILE_LOOPING
        result = all_buffers->check();
        //if( result )
        //    goto cleanup;
#endif

        // give the card drivers a chance to poll their
        // cards for status messages, etc (low bandwidth)
        //
        for( interface=0; interface<MAX_ATM_NICS; interface++ )
            if( cards[interface] )
                cards[interface]->tick();

#if CHECK_ATM_INTERRUPT_ERRORS
        for( interface=0; interface<MAX_ATM_NICS; interface++ )
        {
            result = cards[ interface ]->get_interrupt_result();
            if( result )
                fprintf( stderr, "interface %d saw ATM interrupt error %d\n", interface, result );
        }
#endif

        // if it is time to check for a keystroke...
        //
        now.set();
        if( now >= keystroke_alarm )
        {
            keystroke_alarm = now + keystroke_interval;

            // process the whole keyboard buffer
            //
            while( kbhit() )
            {
                keystroke = getch();
                if( keystroke == 27 )
                    break;

#if DO_FOREGROUND_STATS
                flows_keystroke( keystroke );
#endif
            }

            if( keystroke == 27 )
                break;
        }

#if !ALLOW_TIMER_INTERRUPTS
#if 0
        // poll the RX free and RX compl rings from foreground
        //
        unsigned count2;
        for( count2=0; count2<100; count2++ )
            for( interface=0; interface<MAX_ATM_NICS; interface++ )
                if( cards[interface] )
                    cards[interface]->timer_handler();
#endif
#endif

#if !DO_FOREGROUND_STATS
        // if all the cards have run out of memory...
        //
        if( all_buffers->allocate_failure_count >= MAX_ATM_NICS )

            // stop the program, since nobody
            // will ever free up memory
            //
            break;
#else
#if SIMULATE_FIBER
        block->queried = 0;
#else
#if DO_FULLNESS_STATS
        // update BufferQueue fullness statistics
        //
        count_filled++;
        all_buffers->amount_filled( count );
        sum_filled += count;
        if( max_filled < count )
            max_filled = count;
        if( min_filled > count )
            min_filled = count;
#endif

#if DO_FREENESS_STATS
        // update BufferQueue freeness statistics
        //
        count_free++;
        all_buffers->amount_free( count );
        sum_free += count;
        if( max_free < count )
            max_free = count;
        if( min_free > count )
            min_free = count;
#endif

        // try to get a buffer block
        //
        read_result = all_buffers->read( &block );

        // if we got one...
        //
        if( !read_result )
#endif  // !SIMULATE_FIBER
        {
            block_count++;
            buffer_count += block->used;

#if TRACE_BUFFER_BLOCKS
            printf( "got BufferBlock pointing at 0x%.8X/0x%.8X log/phys\n",
                block->buffers, flat_to_physical(block->buffers) );
            if( 0 && block->buffers == block_trace )
            {
                printf( "saving block to disk..." );
                all_buffers->dump_open( "d:\\badblock.dmp" );
                all_buffers->dump_one( block );
                all_buffers->dump_close();
                printf( " done\n" );
            }
#endif

#if CHECK_MEMORY_BEFORE_PROCESSING
            result = all_buffers->check();
            //if( result )
            //    goto cleanup;
#endif

#if TIME_STATS_CALLS
            before.set();
#endif

#if ALLOW_STATS_CALLS
            // do statistical analysis on it
            //
            result = flows_per_block( block );
            if( result )
                fprintf( stderr, "error %d from full flows_per_block()\n", result );
#else
#if BUSYWAIT_STATS_CALLS
            // busy-wait to simulate foreground CPU usage
            //
            busy_iterations += PentiumClock::wait( (CLOCK_RATE / 1000000) * FLOW_BLOCK_TIME );
#endif
#endif  // ALLOW_STATS_CALLS

#if TIME_STATS_CALLS
            after.set();
#if 0
            elapsed = after - before;
#else
            elapsed = after;
            elapsed -= before;
#endif
            elapsed_tot += elapsed;
            if( elapsed > elapsed_max )
                elapsed_max = elapsed;
            if( elapsed < elapsed_min )
                elapsed_min = elapsed;
#endif  // TIME_STATS_CALLS

#if CHECK_MEMORY_AFTER_PROCESSING
            result = all_buffers->check();
            //if( result )
            //    goto cleanup;
#endif

#if !SIMULATE_FIBER
#if 0
            // this is now the job of cells_delete_cells() in AAL5.C
            //
            // does it leak memory???
            //

            // recycle the buffer
            //
            result = all_buffers->free( block );
            if( result )
            {
                printf( "error %d freeing full buffer block 0x%8.8lX\n", result, block );
                break;
            }
#endif

            // if buffer allocation failed since we last asked...
            //
            if( all_buffers->has_allocate_failed_lately() )

                // for all cards...
                //
                for( interface=0; interface<MAX_ATM_NICS; interface++ )

                    // since cards that fail allocation do not give
                    // further interrupts, give him a shot at this block
                    //
                    cards[interface]->foreground_replenish();
#endif  // !SIMULATE_FIBER
        }

        // if it is time to dump active flows...
        //
        now.set();
        if( now >= active_alarm )
        {
            active_alarm = now + active_interval;

#if ALLOW_STATS_CALLS_2
            // if we are all caught up (the filled list was
            // empty the last time we tried to get a block)...
            //
            if( read_result == ERROR_QUEUE_NOT_FILLED )
                analyze_partial_blocks();
#endif

            /* klt - 10/31/96
             * rifle thru the active list
             * and send it all and delete the
             * queue - this is single-threaded, right?
            printf("DUMPING ACTIVE LIST in main()...\n");
             */
            flows_send_active();
        }

        /* if it is time to check for a query...
         */
        now.set();
        if( now > query_alarm )
        {
            query_alarm = now + query_interval;

            if( flows_is_query_wanted() )
            {
#if ALLOW_STATS_CALLS_2
                // if we are all caught up (the filled list was
                // empty the last time we tried to get a block)...
                //
                if( read_result == ERROR_QUEUE_NOT_FILLED )
                    analyze_partial_blocks();
#endif

#if CHECK_MEMORY_WHILE_QUERYING
                result = all_buffers->check();
                //if( result )
                //    goto cleanup;
#endif

#if DO_FULLNESS_STATS
                // display and clear BufferQueue fullness stats
                //
                sprintf( query_msg+query_msg_len, "BufferQueue fullness %u min, %u max, %f avg over %u iterations\n%n",
                    min_filled, max_filled, (double)sum_filled/count_filled, count_filled, &count );
                max_filled = 0;
                min_filled = BIT32_MAX;
                sum_filled = Bit32( 0 );
                count_filled = 0;
                query_msg_len += count;
#endif

#if DO_FREENESS_STATS
                // display and clear BufferQueue freeness stats
                //
                sprintf( query_msg+query_msg_len, "BufferQueue freeness %u min, %u max, %f avg over %u iterations\n%n",
                    min_free, max_free, (double)sum_free/count_free, count_free, &count );
                max_free = 0;
                min_free = BIT32_MAX;
                sum_free = Bit32( 0 );
                count_free = 0;
                query_msg_len += count;
#endif

                // display and clear BufferQueue allocation failure count
                //
                sprintf( query_msg+query_msg_len, "new BufferQueue allocation failures: %u\n%n",
                    all_buffers->allocate_failure_count-queried_allocate_failures, &count );
                query_msg_len += count;
                queried_allocate_failures = all_buffers->allocate_failure_count;

                // display and clear BufferQueue read_tx failure counts per iface
                //
                for( interface=0; interface<MAX_ATM_NICS; interface++ )
                {
                    sprintf( query_msg+query_msg_len, "BufferQueue::read_tx(%d) failed %d times\n%n",
                        interface,
                        all_buffers->read_tx_failure_count[interface]
                            - queried_read_tx_failures[interface],
                        &count
                    );
                    queried_read_tx_failures[ interface ] =
                        all_buffers->read_tx_failure_count[ interface ];
                    query_msg_len += count;
                }

#if TIME_STATS_CALLS_3
                before.set();
#endif

#if ALLOW_STATS_CALLS_3
                // do summary analysis
                //
                result = flows_query( query_msg );
                if( result )
                {
                    printf( "error %d querying stats\n", result );
                    break;
                }

#else
#if BUSYWAIT_STATS_CALLS_3
                // busy-wait to simulate foreground CPU usage
                //
                PentiumClock::wait( (CLOCK_RATE / 1000) * FLOW_QUERY_TIME );
#endif

                extern unsigned query_count;
                query_count++;
    //          if( query_count % 100 )
    //              printf( " %d", query_count );
    //          else
    //              printf( "capture not doing query %d block %d\n", query_count, block_count );
#endif  // !ALLOW_STATS_CALLS

#if TIME_STATS_CALLS_3 && !defined(DOS)
                after.set();
                elapsed2 = PentiumClock( after - before ).get_seconds() * 1000;
                sprintf( query_msg, "previous flows_query() elapsed time was %g milliseconds\n%n", elapsed2, &query_msg_len );
                printf( query_msg );
#else
                query_msg[0] = '\0';
                query_msg_len = 0;
#endif

                // for all cards...
                //
                for( count=0; count<MAX_ATM_NICS; count++ )

                    // if card transmitter has run dry for this card since we last asked...
                    //
                    // note: assumes that flows_query() might have added some
                    // flows to the sendable list, which is true for PLAYBACK.CPP,
                    // but not for FS2FLOWS.C
                    //
                    if( all_buffers->has_read_tx_failed_lately( count ) )

                        // since cards that fail to get a sendable block do not
                        // give further interrupts, give him a shot at this block
                        //
                        cards[count]->foreground_tx_replenish();
            }   // end if query wanted
        }   // end if query alarm expired
#endif  // DO_FOREGROUND_STATS

    }   // end while forever

#if DO_FOREGROUND_STATS
    // remember what time the main loop exited
    //
    after.set();
#endif

#if CHECK_MEMORY_WHILE_RUNNING
    all_buffers->check();
#endif

cleanup:
    // shut down the cards
    //
    pci_exit();

#if DO_FOREGROUND_STATS
    // calculate how long we were collecting stats
    //
    after -= startup_time;
    printf( "%.2f seconds runtime\n", after.get_seconds() );
    printf( "%.0f buffers and %d blocks seen\n", (double)buffer_count, block_count );
    printf( "so sustained disk write rate needed was %g bytes per second\n", (double)block_count*sizeof(traceblock)/after.get_seconds() );
#endif


#if DO_FOREGROUND_STATS && DO_FULLNESS_STATS
    // display and clear BufferQueue fullness stats
    //
    printf( "BufferQueue fullness %u min, %u max, %f avg over %u iterations\n%n",
        min_filled, max_filled, (double)sum_filled/count_filled, count_filled, &count );
#endif

#if DO_FOREGROUND_STATS && DO_FREENESS_STATS
    // display and clear BufferQueue freeness stats
    //
    printf( "BufferQueue freeness %u min, %u max, %f avg over %u iterations\n%n",
        min_free, max_free, (double)sum_free/count_free, count_free, &count );
#endif

#if DO_FOREGROUND_STATS && (TIME_STATS_CALLS || BUSYWAIT_STATS_CALLS)
    if( block_count )
    {
#if !ALLOW_STATS_CALLS && BUSYWAIT_STATS_CALLS
        printf( "%.4g avg iterations busywaiting %d uS instead of flows_per_block()\n", (double)busy_iterations/block_count, FLOW_BLOCK_TIME );
#endif
#if TIME_STATS_CALLS
        elapsed_min2 = ((double)elapsed_min * 1000) / CLOCK_RATE;
        elapsed_max2 = ((double)elapsed_max * 1000) / CLOCK_RATE;
        elapsed_avg2 = (((double)elapsed_tot / block_count) * 1000) / CLOCK_RATE;
        printf( "flows_per_block() took min %.4f max %.4f avg %.4f mS\n", elapsed_min2, elapsed_max2, elapsed_avg2 );
        printf( "%.2f seconds (%.2f%%) spent in flows_per_block() during run\n", elapsed_tot.get_seconds(), 100.0*(double)elapsed_tot/(double)after );
#endif
    }
    else
        printf( "flows_per_block() was never called\n" );
#endif  // DO_FOREGROUND_STATS && etc.



    printf( "final estimate of card clock rate: %.2f per second\n", CARD_CLOCK_RATE );

#if DEBUG_STACK_OVERFLOW
    printf( "\nstack pointer at top of main() was 0x%4.4X:%8.8lX\n", stack_seg_during_main_top, stack_ptr_during_main_top );
    printf( "stack pointer inside last loop of main() was 0x%4.4X:%8.8lX\n", stack_seg_during_main_loop, stack_ptr_during_main_loop );
    printf( "stack pointer inside first timer_handler() was 0x%4.4X:%8.8lX\n", stack_seg_during_global_timer, stack_ptr_during_global_timer );
#endif

#if DO_FOREGROUND_STATS
    while( 1 )
    {
        // try to get a buffer block
        //
        read_result = all_buffers->read( &block );

        // if we didn't get one, stop
        //
        if( read_result )
            break;

        result = flows_per_block( block );
        if( result )
            fprintf( stderr, "error %d from shutdown flows_per_block()\n", result );

#if 0
        // again, this is the job of cells_delete_cells() in AAL5.C
        // so we cannot free blocks here
        //
        result = all_buffers->free( block );
        if( result )
            fprintf( stderr, "error %d from shutdown block free()\n", result );
#endif
    }

    // clean up stats code and report
    //
    result = flows_cleanup();
#endif

    if( all_buffers )
        delete all_buffers;

    // restore the original interrupt routing
    //
    timer_exit();

    flat_exit();

    // ???maybe force the CMOS clock to be read
    // into the BIOS s/w based real time clock
    //

    return result;
}


