/* conditional compile switches all go here
 */

#define BYTESWAP_FIELDS_EARLY   FALSE       /* must match fs2flows.c */
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define FOLD_BEFORE_MODULUS     TRUE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE       // 12/11/96 1 PM deepest was ~9 at WTN
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_WORKING_SET       TRUE
#define DEBUG_LOCKUPS           FALSE
#define DEBUG                   FALSE
#define SORT_NET_PRINTOUT       FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE       // for debugging when overflow occurs only!  too slow
#define TALK_WHEN_NO_SUBNET     FALSE       // beware: lots of scrolling when TRUE
#define COLLAPSE_FURTHER        FALSE
#define THRESHOLD_AS_COUNTS     FALSE
#define ENABLE_THRESHOLDS       TRUE        // maybe FALSE for vBNS, TRUE for BIPP
#define SHOW_ALL_AGGREGATES     FALSE
#define USE_EXIT_AS             FALSE



#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "../wattcp/tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>             /* errno */
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "../types/types.h"
#   include "../intel/byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"            /* nonzero() variance() */
#include "netstats.h"
#include "readbgp.h"
#include "../dpmi32/flat.h"     /* flat_alloc() */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS


static u_int      src_prefix_flows[ MAX_ADDR_BITS + 1 ];
static Bit64      src_prefix_bytes[ MAX_ADDR_BITS + 1 ];
static u_int    src_prefix_packets[ MAX_ADDR_BITS + 1 ];
static u_int      dst_prefix_flows[ MAX_ADDR_BITS + 1 ];
static Bit64      dst_prefix_bytes[ MAX_ADDR_BITS + 1 ];
static u_int    dst_prefix_packets[ MAX_ADDR_BITS + 1 ];

#define MAX_AS 65536
#define NETREF_POOL_SIZE (MAX_AS * sizeof(int))

static int      *net_ref = NULL;
static int      netthreshold;
static float    cutoffp, cutoffb, cutofff;
static int      netpaircount;



// stuff we need because we do not really allow query
// and clear to occur every time that they are called,
// and instead force them to occur at most every
// NET_POLLING_INTERVAL
//
//#define NET_POLLING_INTERVAL        60.0      // report_net_stats() and init_net_stats() are ignored until this expires
#define NET_POLLING_INTERVAL        3600.0      // report_net_stats() and init_net_stats() are ignored until this expires
#define ISTAMP_UNSET_VALUE     -12345678.0      // cannot use imprecise fractions because we want to test for equality
static PentiumClock net_polling_alarm;
static float istamp, trace_duration;
static u_int flows;
static double packetsum, bytesum, dursum;
static double packet2sum, byte2sum, dur2sum;
static double sumpb, sumbd, sumpd;



#define N_NETPAIRS    65536

#if !FOLD_BEFORE_MODULUS
#   define fold(x) (x)
#   define modulus(x) ((x) % N_NETPAIRS)
#else
#if FOLD_RANDOMIZES
#   if BIT64_AVAILABLE
#       define fold32(x) (((x)>>32) ^ (x))
#       define modulus(x) ((x) & 65535)
#   else
#       define fold32(x) (x)
#       define modulus(x) (x)
#   endif
#   define fold(x) ((fold32(x)*1664525U + 1013904223) >> 16)
#else
#if FOLD_KEEPS_16_BITS
#   define fold(x) (((x)>>16) ^ (x))
#   define modulus(x) ((x) & 65535)
#else
#   define fold(x) (((x)>>24) ^ ((x)>>16) ^ ((x)>>8) ^ (x))
#   define modulus(x) ((x) & 255)
#endif
#endif
#endif

/* generate a 32-bit hash key
 *
 * make sure this uses unsigned shifting
 * by explicitly promoting to Bit32
 */
#if 1
#define nethash(src,dst) modulus( fold( \
                  ((Bit32)src << 16) | dst \
                  ) )
#else
#define nethash(src,dst) modulus( fold( \
                  src + dst \
                  ) )
#endif

#define NETBUCKET_POOL_SIZE (N_NETPAIRS * sizeof(struct nethashrec *))
static struct nethashrec **nettabfull = NULL;
#if COLLAPSE_FURTHER
    static struct nethashrec **nettab = NULL;
#else
#   define nettab nettabfull
#endif

static struct nethashrec *nethash_free = NULL;
static struct nethashrec *nethash_pool = NULL;
u_int nethash_alloc_failures;

// but 145K can be allocated when the BGP table has
// 42000 lines and 60K porthash records are already
// allocated and 350K flow hash records are already
// allocated and DPMIMEM=MAXMEM 49152
//
// oops!  debugger steals exactly 1328 KB before program
// starts, so to allow this program to succeed when run
// under the debugger with all other assumptions unchanged,
// allocate only 128 K records
//
// 12/11/96 BIPP WTN routing table is 96K lines, and refcounts
// must be allocated bigger, so if we still want to run the
// debugger and all other constraints stay the same, can only
// have 117K records
//
// 12/11/96 BIPP WTN actual net hash pool working set
// was about 115K records during a 5-minute period that
// had 600K flows expire, so I am increasing alloc
// 50% to 175K records
//
// 12/11/96 6P changed definition of nethash to use
// source AS number instead of routing table line number,
// so it should be further aggregated now, but idunno again...
//
// 12/12/96 3:15P only needed 19K nethash records for 5 minutes
// where 580K flow hash recs were parsed, so I am reducing
// allocation to 38K
//
// 12/24/96 4:20P forced collection interval to 1 hr, and memory
// might need to increase twelvefold to 456K, but I am allocating
// only 81000, on the theory that the working set will max out
//
// 2/12/97 used up 81K records halfway through each hour,
// so I increased to 150K
//
//#define NETHASH_RECORDS 81000L
#define NETHASH_RECORDS 150000L

#define NETHASH_POOL_SIZE (NETHASH_RECORDS * sizeof(struct nethashrec))

#if DEBUG_CHAIN_LENGTH
#   define NETCHAIN_POOL_SIZE (N_NETPAIRS * sizeof(int))
    static int *chainlen2 = NULL;
    static int longest_chain2;
#if COLLAPSE_FURTHER
    static int *chainlen3 = NULL;
    static int longest_chain3;
#endif
#endif

#if DEBUG_WORKING_SET
    static int total_level2_nethashrecs;
#if COLLAPSE_FURTHER
    static int total_level3_nethashrecs;
#endif
#endif

#if DEBUG_HASHINDEX
    static u_int hashindex_trace = -1;
#endif



#if SORT_NET_PRINTOUT
#define NETPRINT_POOL_SIZE (NETHASH_RECORDS * sizeof(struct nethashrec *))
static Bit32 netcount;
static struct nethashrec **netprintout = NULL;

static int netcmp( void *a, void *b )
{
    struct nethashrec ** hrp1 = (struct nethashrec**) a;
    struct nethashrec ** hrp2 = (struct nethashrec**) b;
    return ((*hrp1)->packets - (*hrp2)->packets);
}
#endif



/*
 * new_net_record
 *      establish a new net pair
 *      1> insert into hash table [maintained as stack]
 */
static
struct nethashrec *
new_net_record( u_short src, u_short dst,
    u_int flows, u_int packets, Bit64 bytes, float duration,
    double p2sum, double b2sum, double d2sum,
    double pbsum, double bdsum, double pdsum
)
{
    struct nethashrec *pr;

    if ((pr = nethash_free) == NULL) {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new nethashrec\n");
#endif
        nethash_alloc_failures++;
        return 0;
    }
    nethash_free = pr->next;

    pr->src=src;
    pr->dst=dst;
    pr->flows=flows;
    pr->packets=packets;
    pr->bytes=bytes;
    pr->duration=duration;
    pr->p2sum=p2sum;
    pr->b2sum=b2sum;
    pr->d2sum=d2sum;
    pr->pbsum=pbsum;
    pr->bdsum=bdsum;
    pr->pdsum=pdsum;
    pr->next=NULL;
    return(pr);
}



static void delete_net_record( struct nethashrec *pr )
{
    /* put this puppy on the free list */
    pr->next = nethash_free;
    nethash_free = pr;
}


/*
 * diagnostic
 */
static void print_net_record( struct nethashrec *pr )
{
    fprintf(stderr, "p: %d ", pr->packets);
    fprintf(stderr, "b: %.0f ", (double)pr->bytes);
    fprintf(stderr, "sdd hash: %d %d %f %d\n",
        pr->src,pr->dst,pr->duration,
        nethash(pr->src,pr->dst));
}



// forgets all statistics
//
int init_net_stats( void )
{
    Bit32 count;
    int result;
    PentiumClock now;

    now.set();

    // if this is the first clear (really initialization)
    //
    if( nethash_pool == NULL )
    {
        fprintf( stderr, "net  hash records are %d bytes each\n", sizeof(struct nethashrec) );
        fprintf( stderr, "radix  trie nodes are %d bytes each\n", sizeof(SubnetTree) );
        fprintf( stderr, "subnet     params are %d bytes each\n", sizeof(Subnet) );

        /* initialize the subnet extraction module
         */
        InitSubnet();

        /* parse the main BGP table into a radix trie that points at subnets
         */
        result = read_bgp_file( BGP_PATH "bgp.txt" );
        if( result )
            return result;

        /* parse the exceptions BGP table into the same radix trie
         */
        result = read_bgp_file( BGP_PATH "bgp-exc.txt" );
        if( result )
            return result;

#if DEBUG_CHAIN_LENGTH
        /* allocate chain length counters for each bucket in main hash table */
        chainlen2 = (int *) flat_alloc( NETCHAIN_POOL_SIZE );
        if( chainlen2 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for main net chain length counters\n", errno, NETCHAIN_POOL_SIZE, N_NETPAIRS );
            return 1;
        }

#if COLLAPSE_FURTHER
        /* allocate chain length counters for each bucket in secondary hash table */
        chainlen3 = (int *) flat_alloc( NETCHAIN_POOL_SIZE );
        if( chainlen3 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for secondary net chain length counters\n", errno, NETCHAIN_POOL_SIZE, N_NETPAIRS );
            return 1;
        }
#endif
#endif

        // allocate the AS reference counters
        //
        net_ref = (int *) flat_alloc( NETREF_POOL_SIZE );
        if( net_ref == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld integers) for AS reference counters\n", errno, NETREF_POOL_SIZE, MAX_AS );
            return 1;
        }

        /* allocate the buckets for the main hash table */
        nettabfull = (struct nethashrec **) flat_alloc( NETBUCKET_POOL_SIZE );
        if( nettabfull == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for main port hash buckets\n", errno, NETBUCKET_POOL_SIZE, N_NETPAIRS );
            return 1;
        }

        /* allocate the buckets for the secondary hash table */
        nettab = (struct nethashrec **) flat_alloc( NETBUCKET_POOL_SIZE );
        if( nettab == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for secondary port hash buckets\n", errno, NETBUCKET_POOL_SIZE, N_NETPAIRS );
            return 1;
        }

#if SORT_NET_PRINTOUT
        /* allocate pointers to sort during report */
        netprintout = (struct nethashrec **) flat_alloc( NETPRINT_POOL_SIZE );
        if( netprintout == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for net printout sorting\n", errno, NETPRINT_POOL_SIZE, NETHASH_RECORDS );
            return 1;
        }
#endif

        /* allocate a pool of hash records */
        nethash_pool = (struct nethashrec *) flat_alloc( NETHASH_POOL_SIZE );
        if( nethash_pool == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld records) for net hash record pool\n", errno, NETHASH_POOL_SIZE, NETHASH_RECORDS );
            return 1;
        }

        /* put all the hash records on the singly-linked free list */
        nethash_free = nethash_pool;
        for( count=0; count<NETHASH_RECORDS-1; count++ )
        {
            nethash_pool[ count ].next = nethash_pool + count + 1;
        }
        nethash_pool[count].next = NULL;
    }
    else
    {
        // if not enough time has gone by since last polling...
        //
        if( now < net_polling_alarm )

            // don't allow clear to occur
            //
            return 0;
    }

    // reset the polling-allowed alarm
    //
    net_polling_alarm.put_seconds( NET_POLLING_INTERVAL );
    net_polling_alarm += now;

    // no pollings have occurred since one that was allowed
    // (assumes that valid istamp will never have this value)
    //
    istamp = ISTAMP_UNSET_VALUE;

    // reset the accumulators of report_net_stats() parameters that
    // we have to keep because we are forcing the polling interval
    //
    trace_duration  = 0.0;
    flows           = 0.0;
    packetsum       = 0.0;
    bytesum         = 0.0;
    dursum          = 0.0;
    packet2sum      = 0.0;
    byte2sum        = 0.0;
    dur2sum         = 0.0;
    sumpb           = 0.0;
    sumbd           = 0.0;
    sumpd           = 0.0;

    // no prefix lengths have been referenced yet
    //
    for( count=0; count<MAX_ADDR_BITS+1; count++ )
    {
        src_prefix_flows[ count ] = 0;
        src_prefix_packets[ count ] = 0;
        src_prefix_bytes[ count ] = (Bit32)0;
        dst_prefix_flows[ count ] = 0;
        dst_prefix_packets[ count ] = 0;
        dst_prefix_bytes[ count ] = (Bit32)0;
    }

    // no nets have been referenced yet
    //
    for( count=0; count<MAX_AS; count++ )
    {
        net_ref[count]=0;
    }

    for( count=0; count<N_NETPAIRS; count++ )
    {
        // level2 hashtable is empty
        //
        nettabfull[count] = 0;

#if COLLAPSE_FURTHER
        // level3 hashtable is empty
        //
        nettab[count] = 0;
#endif

#if SORT_NET_PRINTOUT
        // no net pairs are ready to print
        //
        netprintout[ count ] = 0;
#endif

#if DEBUG_CHAIN_LENGTH
        /* initialize chain length of all hashtable buckets to zero
         */
        chainlen2[ count ] = 0;
#if COLLAPSE_FURTHER
        chainlen3[ count ] = 0;
#endif
#endif
    }

#if ENABLE_THRESHOLDS
    cutoffp = .0005;
    cutoffb = .0005;
    cutofff = .0005;
#else
    cutoffp = 0.0;
    cutoffb = 0.0;
    cutofff = 0.0;
#endif
#if THRESHOLD_AS_COUNTS
    netthreshold = 5;
#else
    netthreshold = 0;
#endif

    netpaircount = 0;

#if DEBUG_CHAIN_LENGTH
    longest_chain2 = 0;
#if COLLAPSE_FURTHER
    longest_chain3 = 0;
#endif
#endif

#if DEBUG_WORKING_SET
    total_level2_nethashrecs = 0;
#if COLLAPSE_FURTHER
    total_level3_nethashrecs = 0;
#endif
#endif

    nethash_alloc_failures = 0;

    return 0;
}



void update_net_stats( struct hashrec *hr,
    u_int packets, Bit64 bytes, float duration,
    double packets2, double bytes2, double duration2,
    double packets_bytes, double bytes_duration, double packets_duration
)
{
        unsigned nethashindex;
        int cmp;
        struct nethashrec *pp, *neth, *newpp;
        Subnet *src_subnet, *dst_subnet;
        u_short src, dst;
        unsigned prefix_length;

#if BYTESWAP_FIELDS_EARLY
        src_subnet = FindSubnet( htonl( hr->hr_fd.fd_src.s_addr ) );
        dst_subnet = FindSubnet( htonl( hr->hr_fd.fd_dst.s_addr ) );
#else
        src_subnet = FindSubnet( hr->hr_fd.fd_src.s_addr );
        dst_subnet = FindSubnet( hr->hr_fd.fd_dst.s_addr );
#endif

        if( src_subnet )
        {
#if USE_EXIT_AS
            src = src_subnet->exit_as;
#else
            src = src_subnet->src_as;
#endif
            prefix_length = bits_in_mask( src_subnet->mask );
            src_prefix_flows[ prefix_length ]++;
            src_prefix_bytes[ prefix_length ] += bytes;
            src_prefix_packets[ prefix_length ] += packets;
        }
        else
        {
#if TALK_WHEN_NO_SUBNET
            char temp[ 20 ];

            fprintf( stderr, "update_net_stats() warning: cannot find subnet for source IP %s\n",
#if BYTESWAP_FIELDS_EARLY
                inet_ntoa( temp, hr->hr_fd.fd_src.s_addr )
#else
                inet_ntoa( temp, ntohl( hr->hr_fd.fd_src.s_addr ) )
#endif
                );
#endif

            src = UNFOUND_SRC_AS;
        }

        if( dst_subnet )
        {
#if USE_EXIT_AS
            dst = dst_subnet->exit_as;
#else
            dst = dst_subnet->src_as;
#endif
            prefix_length = bits_in_mask( dst_subnet->mask );
            dst_prefix_flows[ prefix_length ]++;
            dst_prefix_bytes[ prefix_length ] += bytes;
            dst_prefix_packets[ prefix_length ] += packets;
        }
        else
        {
#if TALK_WHEN_NO_SUBNET
            char temp[ 20 ];

            fprintf( stderr, "update_net_stats() warning: cannot find subnet for destination IP %s\n",
#if BYTESWAP_FIELDS_EARLY
                inet_ntoa( temp, hr->hr_fd.fd_dst.s_addr )
#else
                inet_ntoa( temp, ntohl( hr->hr_fd.fd_dst.s_addr ) )
#endif
                );
#endif

            dst = UNFOUND_DST_AS;
        }

        /* update ref counts for AS numbers
         * for both source and dest AS
         */
        net_ref[ src ]++;
        net_ref[ dst ]++;

        nethashindex = nethash(src,dst);
        pp = nettabfull[nethashindex];
        /* look for match or collision on hash chain */
        neth=pp;
        cmp=0;
        while (neth!=NULL && cmp==0)
        {
                cmp = (src==neth->src && dst==neth->dst);
                if (cmp) {
                        /* update all the stats in nethashrec */
                        neth->flows++;
                        neth->packets+=packets;
                        neth->bytes+=bytes;
                        neth->duration+=duration;
                        neth->p2sum+=packets2;
                        neth->b2sum+=bytes2;
                        neth->d2sum+=duration2;
                        neth->pbsum+=packets_bytes;
                        neth->bdsum+=bytes_duration;
                        neth->pdsum+=packets_duration;
#if DEBUG
                        /* for debugging */
                        printf("match\t");
                        print_net_record(neth);
#endif
                }
                neth=neth->next;
        };

        /* didn't find it anywhere, must be a new record */
        if (cmp==0)
        {
            netpaircount++;
            newpp = new_net_record( src, dst, 1, packets, bytes, duration,
                packets2, bytes2, duration2, packets_bytes, bytes_duration,
                packets_duration );
            if( newpp )
            {
                newpp->next = pp;
                nettabfull[nethashindex] = newpp;
#if DEBUG_WORKING_SET
                /* increment count of level2 hash records
                 */
                total_level2_nethashrecs++;
#endif

#if DEBUG_CHAIN_LENGTH
                /*
                 * increment chain length for this hash index
                 */
                chainlen2[nethashindex]++;

#if DEBUG_HASHINDEX
                if( nethashindex == hashindex_trace )
                    fprintf( stderr, "at packet %d hash index 0x%8.8lX chain length increased to %d\n", total_packets, nethashindex, chainlen2[nethashindex] );
#endif

                /*
                 * if this is the longest chain we have ever seen,
                 * update our global, and notify the user
                 *
                 * notifications should be asymptotically decreasing
                 */
                if( longest_chain2 < chainlen2[nethashindex] )
                {
                    longest_chain2 = chainlen2[nethashindex];
                    fprintf( stderr, "hash index 0x%8.8lX is now longest net chain 2 at %d\n", nethashindex, longest_chain2 );
                    fprintf( stderr, "calced from subnet src=%d dst=%d\n", src, dst );
                }
#endif

/* for debugging */
#if DEBUG
                printf("new\t");
                print_net_record(newpp);
#endif
            }   // end if could alloc nethashrec
        }   // end if needed a new nethashrec
}



/* "SNB src_as dst_as flows packets bytes dur %%f %%p %%b %%d fps "
   "pps bps ppf bpf spf bpp vp vb vd copb cobd copd \n");
*/
static void print_net_stats( tcp_Socket *sock, float trace_duration,
    int total_flows, struct nethashrec *neth, double packetsum,
    double bytesum, double dursum )
{
    sock_printf(
        sock,
        "SNB\t%d\t%d\t%d\t%d\t%.0f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.0f\t%.0f\t"
            "%.3f\t%.0f\t%.3f\t%.0f\t"
            "%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.3f\t%.3f\n",

        neth->src,
        neth->dst,

        neth->flows,
        neth->packets,
        (double) neth->bytes,
        neth->duration,

        (float) neth->flows/total_flows,
        (float) neth->packets/packetsum,
        (float) neth->bytes/bytesum,
        (float) neth->duration/dursum,

        neth->flows/nonzero(trace_duration),
        neth->packets/nonzero(trace_duration),
        neth->bytes/nonzero(trace_duration),

        (float) neth->packets/neth->flows,
        (float) neth->bytes/neth->flows,
        (float) neth->duration/neth->flows,

        (float) neth->bytes/neth->packets,

        variance(neth->flows,neth->packets,neth->p2sum),
        variance(neth->flows,neth->bytes,neth->b2sum),
        variance(neth->flows,neth->duration,neth->d2sum),

        (float) neth->pbsum,
        (float) neth->bdsum,
        (float) neth->pdsum                // 23rd field
    );
}



int report_net_stats( tcp_Socket *sock,
    float istamp0, float trace_duration0, u_int flows0,
    double packetsum0, double bytesum0, double dursum0,
    double packet2sum0, double byte2sum0, double dur2sum0,
    double sumpb0, double sumbd0, double sumpd0 )
{
        int i, expcount=0, exprefcount=0;
        unsigned nethashindex;
        int cmp;
        struct nethashrec *pp, *nethfull, *neth, *newpp, *next;
        PentiumClock now;
        u_int ignored_flows = 0;
        u_int ignored_packets = 0;
        Bit64 ignored_bytes = (Bit32) 0;
        double ignored_duration = 0.0;
        Bit64 cutoffb_bit64;
        const char *aggregate_as_string;
        unsigned prefix_length;

        // if this is the first polling after one we allowed...
        //
        if( istamp == ISTAMP_UNSET_VALUE )

            // remember the initial timestamp
            //
            istamp = istamp0;

        // accumulate the parameters that we will need for report
        // when we finally allow the polling to proceed
        //
        trace_duration  += trace_duration0;
        flows           += flows0;
        packetsum       += packetsum0;
        bytesum         += bytesum0;
        dursum          += dursum0;
        packet2sum      += packet2sum0;
        byte2sum        += byte2sum0;
        dur2sum         += dur2sum0;
        sumpb           += sumpb0;
        sumbd           += sumbd0;
        sumpd           += sumpd0;

        // print out a summary since we last allowed the polling
        // to proceed, since the ST line output by report_stats()
        // at the top of the report only applies to the current
        // polling attempt
        //
        sock_printf (sock, "SNT\t%.0f\t%.3f\t%d\t%.f\t%.lf\t%.0f\t%.0f\t%.0f\t%.0f\t"
                                "1.0\t1.0\t1.0\t1.0\t"
                                "%.3f\t%.0f\t%.0f\t"
                                "%.3f\t%.0f\t%.3f\t%.0f\t"
                                "%.3f\t%.3f\t%.3f\n",
                istamp,
                trace_duration,

                flows,
                packetsum,
                bytesum,
                dursum,

                flows/nonzero(trace_duration),
                packetsum/nonzero(trace_duration),
                bytesum/nonzero(trace_duration),

                (float) packetsum/nonzero(flows),
                (float) bytesum/nonzero(flows),
                dursum/nonzero(flows),

                (float) bytesum/nonzero(packetsum),

                variance(flows,packetsum,packet2sum),
                variance(flows,bytesum,byte2sum),
                variance(flows,dursum,dur2sum),

                sumpb,
                sumbd,
                sumpd);

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 2 net hash records created\n", total_level2_nethashrecs );
#endif

        sock_printf( sock, "%d/%d source/destination net lookup failures\n", net_ref[UNFOUND_SRC_AS], net_ref[UNFOUND_DST_AS] );

#if DEBUG_LOCKUPS
        fprintf( stderr, "SNS" );
#endif
        /* Display nets with excessive references in pairings.  */
        /*  SNS net1:ref1 net2:ref2 net3:ref3 */
        sock_printf(sock,"SNS\t");
        for (i=0;i<MAX_AS;i++) {
            if (net_ref[i] > netthreshold)
            {
                sock_printf(sock,"%d:%d\t",i,net_ref[i]);
                expcount++;
                exprefcount = exprefcount + net_ref[i];
            }
        };
        sock_printf(sock,"\n");

#if DEBUG_LOCKUPS
        fprintf( stderr, " SNC" );
#endif
        /* SNC refs-threshold nets-that-matched total-matching-refs total-net-pairs  */
        sock_printf(sock,"SNC\t%d\t%d\t%d\t%d\n",
                netthreshold,expcount,exprefcount,netpaircount);

        /* display mappings of fake AS numbers to aggregates
         */
        for( i=AGGREGATE_AS_BASE; i<=AGGREGATE_AS_MAX; i++ )
        {
            aggregate_as_string = get_aggregate_as_expansion( i );
            if( aggregate_as_string == NULL )           // last one?
                break;

#if !SHOW_ALL_AGGREGATES
            if( net_ref[i] > netthreshold )
#endif
                sock_printf( sock, "SNM %d='%s'\n", i, aggregate_as_string );
        }

        /* display flow/packet/byte counts for prefix lengths
         * for src/dst IP addresses
         */
        sock_printf( sock, "SNLSF" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( src_prefix_flows[ prefix_length ] )
                sock_printf( sock, " %u:%u", prefix_length, src_prefix_flows[prefix_length] );

        sock_printf( sock, "\nSNLSP" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( src_prefix_packets[ prefix_length ] )
                sock_printf( sock, " %u:%u", prefix_length, src_prefix_packets[prefix_length] );

        sock_printf( sock, "\nSNLSB" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( src_prefix_bytes[ prefix_length ] )
                sock_printf( sock, " %u:%.0f", prefix_length, (double)src_prefix_bytes[prefix_length] );

        sock_printf( sock, "\nSNLDF" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( dst_prefix_flows[ prefix_length ] )
                sock_printf( sock, " %u:%u", prefix_length, dst_prefix_flows[prefix_length] );

        sock_printf( sock, "\nSNLDP" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( dst_prefix_packets[ prefix_length ] )
                sock_printf( sock, " %u:%u", prefix_length, dst_prefix_packets[prefix_length] );

        sock_printf( sock, "\nSNLDB" );
        for( prefix_length=0; prefix_length<MAX_ADDR_BITS+1; prefix_length++ )
            if( dst_prefix_bytes[ prefix_length ] )
                sock_printf( sock, " %u:%.0f", prefix_length, (double)dst_prefix_bytes[prefix_length] );
        sock_printf( sock, "\n" );

        // if not enough time has gone by since last polling...
        //
        now.set();
        if( now < net_polling_alarm )
        {
            sock_printf( sock, "net pair reporting/clearing disabled for %.1f more seconds\n",
                PentiumClock(net_polling_alarm - now).get_seconds() );

#if DEBUG_LOCKUPS
            fprintf( stderr, "\n" );
#endif
            // don't output any SNB lines
            //
            return 0;
        }

#if COLLAPSE_FURTHER
#if DEBUG_LOCKUPS
        fprintf( stderr, " collapse" );
#endif
        /* collapse further
         *
         * we could not do this on the first pass because
         * we did not know (total usage per net) yet
         */
        for (i=0;i<N_NETPAIRS;i++) {
            for (nethfull=nettabfull[i]; nethfull!=NULL; nethfull=next)
            {
                next = nethfull->next;

                // if source AS had more than the threshold flows
                // and source AS is not already collapsed
                // and source AS had more flows than dest AS...
                //
                if ((net_ref[nethfull->src] > netthreshold) &&
                    nethfull->src != COLLAPSED_AS &&
                    (net_ref[nethfull->dst] < net_ref[nethfull->src]))
                {
                    nethfull->dst = COLLAPSED_AS;
                }
                else
                {
                    if ((net_ref[nethfull->dst] > netthreshold) &&
                        nethfull->dst != COLLAPSED_AS &&
                        (net_ref[nethfull->src] < net_ref[nethfull->dst]))
                    {
                        nethfull->src = COLLAPSED_AS;
                    }
                }

                /* rehash into new array */
                nethashindex = nethash(nethfull->src,nethfull->dst);
                pp = nettab[nethashindex];
                neth=pp;
                cmp=0;
                while (neth != NULL && cmp==0) {
/* for debugging */
#if DEBUG
                    fprintf(stderr,"seeking match netab\t");
                    print_net_record(neth);
#endif
                    cmp = ( (nethfull->src==neth->src)
                        &&  (nethfull->dst==neth->dst) );
                    if (cmp) {
                        neth->flows+=nethfull->flows;
                        neth->packets+=nethfull->packets;
                        neth->bytes+=nethfull->bytes;
                        neth->duration+=nethfull->duration;
                        neth->p2sum+=nethfull->p2sum;
                        neth->b2sum+=nethfull->b2sum;
                        neth->d2sum+=nethfull->d2sum;
                        neth->pbsum+=nethfull->pbsum;
                        neth->bdsum+=nethfull->bdsum;
                        neth->pdsum+=nethfull->pdsum;
    /* for debugging */
#if DEBUG
                        fprintf(stderr,"match netab\t");
                        print_net_record(neth);
#endif
                    }
                    neth=neth->next;

                }   /* end search hash chain 3 */

/* didn't find it anywhere, must be a new record */
                if (cmp == 0)
                {
                    newpp = new_net_record(
                        nethfull->src, nethfull->dst, nethfull->flows,
                        nethfull->packets, nethfull->bytes, nethfull->duration,
                        nethfull->p2sum, nethfull->b2sum, nethfull->d2sum,
                        nethfull->pbsum, nethfull->bdsum, nethfull->pdsum );
                    if( newpp )
                    {
                        newpp->next = pp;
                        nettab[nethashindex] = newpp;
#if DEBUG_WORKING_SET
                        /* increment count of level3 hash records
                         */
                        total_level3_nethashrecs++;
#endif
#if DEBUG_CHAIN_LENGTH
                        /*
                         * increment chain length for this hash index
                         */
                        chainlen3[nethashindex]++;

                        /*
                         * if this is the longest chain we have ever seen,
                         * update our global, and notify the user
                         *
                         * notifications should be asymptotically decreasing
                         */
                        if( longest_chain3 < chainlen3[nethashindex] )
                        {
                            longest_chain3 = chainlen3[nethashindex];
                            fprintf( stderr, "hash index 0x%8.8lX is now longest net chain 3 at %d\n", nethashindex, longest_chain3 );
                            fprintf( stderr, "calced from subnet src=%d dst=%d\n", nethfull->src, nethfull->dst );
                        }
#endif
                    }   // end if could alloc nethashrec
                }

                /* free up the nettabfull entry (level 2 hash)
                 * so code can run continuously...
                 */
                delete_net_record( nethfull );

            }   /* end search hash chain 2 */
        }   /* end for all level 2 chains */

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 3 net hash records created\n", total_level3_nethashrecs );
#endif
#endif  // COLLAPSE_FURTHER

        /* go through nettab and print  now */
        cutoffp *= packetsum;
        cutoffb *= bytesum;
        cutoffb_bit64 = cutoffb;
        cutofff *= flows;

#if DEBUG_LOCKUPS
        fprintf( stderr, " SNB" );
#endif
        for (i=0;i<N_NETPAIRS;i++) {
            for (neth=nettab[i]; neth!=NULL; neth=next) {
                next = neth->next;

                if ((neth->packets > cutoffp) ||
                    (neth->bytes > cutoffb_bit64) ||
                    (neth->flows > cutofff))
                {
#if SORT_NET_PRINTOUT
                    /* this will be deleted after sort/print */
                    netcount++;
                    netprintout[netcount]=neth;
#else
                    print_net_stats(sock, trace_duration, flows, neth, packetsum, bytesum, dursum);

                    /* we are already done with this record */
                    delete_net_record( neth );
#endif
                }
                else
                {
                    /* count how much I am not telling you
                     */
                    ignored_flows    += neth->flows;
                    ignored_packets  += neth->packets;
                    ignored_bytes    += neth->bytes;
                    ignored_duration += neth->duration;

                    /* done with this net pair regardless whether
                     * output will be sorted, since it isn't in
                     * the output
                     */
                    delete_net_record( neth );
                }
            }
        }

#if SORT_NET_PRINTOUT
        /* sort from nethle array 80 each */
        qsort( netprintout, netcount, sizeof(struct nethashrec *), &netcmp );
        for( i=0; i<netcount; i++)
        {
            print_net_stats( sock, trace_duration, netprintout[i],packetsum, bytesum, dursum );
            delete_net_record( netprintout[i] );
        }
#endif

        sock_printf( sock, "%u flows, %u packets, %.0f bytes, %.1f duration ignored by network report\n",
            ignored_flows, ignored_packets, (double)ignored_bytes, ignored_duration );

#if DEBUG_LOCKUPS
        fprintf( stderr, "\n" );
#endif
        return 0;
}



#endif  // DO_FOREGROUND_STATS
